import heapq

import numpy as np


def dijkstra_distances(src, nv, neighbors, neighbor_distances, max_dist=None):
    """Compute the Dijkstra distances from one vertex to all others.

    Parameters
    ----------
    src : int
        The index of the source vertex. The computed distances are distances
        to this vertex.
    nv : int
        The total number of vertices.
    neighbors : list of ndarray
        Each element in the list is an ndarray of integers, which are the
        indices of the neighbors of a vertex.
    neighbor_distances : list of ndarray
        Each element in the list is an ndarray of floats, which are the
        distances between the vertex and its neighbors. The order is the same
        as ``neighbors``.
    max_dist : {float, None}, default=None
        The maximal distance. The algorithm will stop when ``max_dist`` is
        reached, and the distances more than ``max_dist`` will be filled with
        ``np.inf``. If it's None, there will be no limit, and the algorithm
        will only stop when all Dijkstra distances have been computed.

    Returns
    -------
    dists : ndarray of shape (nv, )
        The Dijkstra distances between the source vertex and all vertices.
    """
    dists = np.full((nv,), np.inf)
    finished = np.zeros((nv,), dtype=bool)
    candidates = [(0.0, src)]
    dists[src] = 0.0

    while candidates:
        d, idx = heapq.heappop(candidates)
        if finished[idx]:
            continue

        for nbr, nbr_d in zip(neighbors[idx], neighbor_distances[idx]):
            new_d = d + nbr_d
            if max_dist is not None and new_d > max_dist:
                continue
            if new_d < dists[nbr]:
                dists[nbr] = new_d
                heapq.heappush(candidates, (new_d, nbr))
        finished[idx] = True

    return dists


def dijkstra(src, nv, neighbors, neighbor_distances, max_dist, sort=True):
    """Compute the Dijkstra neighbors and distances to these neighbors.

    Parameters
    ----------
    src : int
        The index of the source vertex. The computed distances are distances
        to this vertex.
    nv : int
        The total number of vertices.
    neighbors : list of ndarray
        Each element in the list is an ndarray of integers, which are the
        indices of the neighbors of a vertex.
    neighbor_distances : list of ndarray
        Each element in the list is an ndarray of floats, which are the
        distances between the vertex and the neighbors. The order is the same
        as ``neighbors``.
    max_dist : float
        The maximal distance. The algorithm will stop when ``max_dist`` is
        reached.
    sort : bool, default=True
        Whether to sort the Dijkstra neighbors by the distance (nearest
        first).

    Returns
    -------
    dijkstra_nbrs : ndarray
        The indices of vertices that are within ``max_dist`` of the source
        vertex based on Dijkstra distance.
    dijkstra_dists : ndarray
        The Dijkstra distances.
    """
    dijkstra_dists = dijkstra_distances(
        src, nv, neighbors, neighbor_distances, max_dist
    )
    mask = dijkstra_dists < max_dist
    dijkstra_nbrs = np.where(mask)[0]
    dijkstra_dists = dijkstra_dists[mask]
    if sort:
        sort_idx = np.argsort(dijkstra_dists)
        dijkstra_nbrs = dijkstra_nbrs[sort_idx]
        dijkstra_dists = dijkstra_dists[sort_idx]
    return dijkstra_nbrs, dijkstra_dists



import os

import numpy as np


def triangle_subdivision(n_div):
    faces = []
    for i in range(1, n_div + 1)[::-1]:
        for j in range(n_div + 1 - i):
            k = n_div - i - j
            f = [(i, j, k), (i - 1, j + 1, k), (i - 1, j, k + 1)]
            faces.append(f)
    for i in range(1, n_div + 1)[::-1]:
        for j in range(1, n_div + 1 - i):
            k = n_div - i - j
            f = [(i, j, k), (i - 1, j, k + 1), (i, j - 1, k + 1)]
            faces.append(f)
    return faces


def subdivide_edges(coords, faces, n_div):
    n_edges = faces.shape[0] * 3 // 2
    nv_new = n_edges * (n_div - 1)
    new_coords = np.zeros((nv_new, 3), dtype=coords.dtype)

    nv = coords.shape[0]
    count = 0
    edges = set()
    e_mapping = {}
    for f in faces:
        for a, b in [[f[0], f[1]], [f[0], f[2]], [f[1], f[2]]]:
            e = (a, b) if a < b else (b, a)
            if e in edges:
                continue
            edges.add(e)
            for i in range(1, n_div):
                c = (coords[a] * i + coords[b] * (n_div - i)) / n_div
                # c /= np.linalg.norm(c)
                # new_coords.append(c)
                new_coords[count] = c
                e_mapping[(e[0], e[1], i)] = count + nv
                count += 1

    return new_coords, e_mapping


def subdivide_inside(coords, faces, e_mapping, n_div, count_base):
    # new_coords, new_faces = [], []
    nv_new = (n_div - 1) * (n_div - 2) // 2 * len(faces)
    new_coords = np.zeros((nv_new, 3), dtype=coords.dtype)
    count = 0
    new_faces = []
    nf_base = triangle_subdivision(n_div)

    for f in faces:
        mapping = {
            (n_div, 0, 0): f[0],
            (0, n_div, 0): f[1],
            (0, 0, n_div): f[2],
        }

        for ii, jj in [[0, 1], [0, 2], [1, 2]]:
            a, b = f[[ii, jj]]
            aa, bb = (a, b) if a < b else (b, a)
            for step in range(1, n_div):
                val = e_mapping[aa, bb, step]
                key = [0, 0, 0]
                key[ii] = step
                key[jj] = n_div - step
                mapping[tuple(key)] = val

        for i in range(n_div)[::-1]:
            for j in range(n_div + 1 - i):
                k = n_div - i - j
                if (i, j, k) not in mapping:
                    mapping[(i, j, k)] = count + count_base
                    c = np.sum(
                        coords[f] * (np.array([i, j, k])[:, np.newaxis] / n_div), axis=0
                    )
                    # c /= np.linalg.norm(c)
                    # new_coords.append(c)
                    new_coords[count] = c
                    count += 1

        nf = [[mapping[v] for v in f] for f in nf_base]
        new_faces += nf

    # new_coords = np.array(new_coords)
    new_faces = np.array(new_faces)
    return new_coords, new_faces


def surface_subdivision(coords, faces, n_div):
    edge_coords, e_mapping = subdivide_edges(coords, faces, n_div)
    count = coords.shape[0] + edge_coords.shape[0]
    inside_coords, new_faces = subdivide_inside(coords, faces, e_mapping, n_div, count)
    print(coords.shape, edge_coords.shape, inside_coords.shape)
    new_coords = np.concatenate([coords, edge_coords, inside_coords], axis=0)
    return new_coords, new_faces


#     new_coords, new_faces = [], []

#     # count = coords.shape[0]

#     new_coords_edges, e_mapping = subdivide_edges(coords, faces, n_div)


#     new_coords = np.array(new_coords)
#     new_faces = np.array(new_faces)
#     coords = np.concatenate([coords, new_coords], axis=0)

#     return coords, new_faces



import os

import numpy as np
import scipy.sparse as sparse

from .io import core_dataset

MEASURES = [
    'area',
    'area.mid',
    'area.pial',
    'curv',
    'curv.pial',
    'jacobian_white',
    'sulc',
    'thickness',
    'volume',
]
PARCELLATIONS = ['aparc', 'aparc.DKTatlas', 'aparc.a2009s']


def get_morphometry(which, lr, space='onavg-ico32', **kwargs):
    """Group-based morphometry measure.

    Parameters
    ----------
    which : str
        Which morphometric measure to get. One of the following:
        'area', 'area.mid', 'area.pial', 'curv', 'curv.pial',
        'jacobian_white', 'sulc', 'thickness', 'volume'.
    lr : str
        Hemisphere, either 'l' or 'r'.
    space : str, default='onavg-ico32'
        Surface space.

    Returns
    -------
    measure : ndarray
        Morphometric measure. Shape (n_vertices,).
    """
    assert which in MEASURES
    group = kwargs.get('group', 'on1031')
    resample = kwargs.get('resample', 'overlap-8div')
    avg_type = kwargs.get('avg_type', 'trimmed')
    assert lr in 'lr'
    fn = os.path.join(
        space, 'morphometry', which, f'{lr}h', f'{group}_{avg_type}', f'{resample}.npy'
    )
    measure = core_dataset.get(fn, on_missing='raise')
    return measure


def get_parcellation(which, lr, space='onavg-ico32', prob=False, **kwargs):
    """Group-based parcellation.

    Parameters
    ----------
    which : str
        Which parcellation to get. One of the following:
        'aparc', 'aparc.DKTatlas', 'aparc.a2009s'.
    lr : str
        Hemisphere, either 'l' or 'r'.
    space : str, default='onavg-ico32'
        Surface space.
    prob : bool, default=False
        Whether to load the probabilistic version of the parcellation.

    Returns
    -------
    parc : ndarray
        The cortical parcellation. It is a discrete parcellation if ``prob``
        is False, and a probabilistic parcellation if ``prob`` is True.
    """
    assert which in PARCELLATIONS
    group = kwargs.get('group', 'on1031')
    resample = kwargs.get('resample', 'overlap-8div')
    avg_type = kwargs.get('avg_type', 'trimmed')
    assert lr in 'lr'
    if prob:
        basename = f'{resample}_prob.npy'
    else:
        basename = f'{resample}_parc.npy'
    fn = os.path.join(
        space, 'parcellations', which, f'{lr}h', f'{group}_{avg_type}', basename
    )
    parc = core_dataset.get(fn, on_missing='raise')
    return parc


def get_mask(lr, space='onavg-ico32', legacy=False, **kwargs):
    """Standard cortical mask.

    Parameters
    ----------
    lr : str
        Hemisphere, either 'l' or 'r'.
    space : str, default='onavg-ico32'
        Surface space.
    legacy : bool, default=False
        Whether to load the legacy version of the mask, e.g., "fsaverage" or
        "32k_fs_LR", instead of the group-based cortical mask.

    Returns
    -------
    mask : ndarray
        Mask of the cortical surface. Shape (n_vertices,).
    """
    resample = kwargs.get('resample', 'overlap-8div')
    which = kwargs.get('which', 'aparc.a2009s')
    assert lr in 'lr'
    if legacy:
        if space.startswith('fsavg-ico'):
            flavor = kwargs.get('flavor', 'fsaverage')
        elif space.startswith('fslr-ico'):
            flavor = kwargs.get('flavor', '32k_fs_LR')
        fn = os.path.join(
            space, 'masks', which, f'{lr}h', f'{flavor}', f'{resample}.npy'
        )
    else:
        group = kwargs.get('group', 'on1031')
        avg_type = kwargs.get('avg_type', 'trimmed')
        fn = os.path.join(
            space, 'masks', which, f'{lr}h', f'{group}_{avg_type}', f'{resample}.npy'
        )
    mask = core_dataset.get(fn, on_missing='raise')
    return mask


def get_geometry(which, lr, space='onavg-ico32', vertices_only=False, **kwargs):
    """Surface geometry.

    Parameters
    ----------
    which : str
        Which geometry to get. One of the following:
        'sphere', 'sphere.reg', 'white', 'pial', 'inflated', 'midthickness',
        'faces'.
    lr : str
        Hemisphere, either 'l' or 'r'.
    space : str, default='onavg-ico32'
        Surface space.
    vertices_only : bool, default=False
        Whether to return only the coordinates of the vertices.

    Returns
    -------
    coords : ndarray
        Coordinates of the vertices. Shape (n_vertices, 3).
        When ``which == 'faces'``, ``coords`` will not be returned.
    faces : ndarray
        Faces of the triangle surface mesh. Shape (n_faces, 3).
        Only returned when ``vertices_only == False``.
    """
    group = kwargs.get('group', 'on1031')
    avg_type = kwargs.get('avg_type', 'trimmed')
    assert lr in 'lr'
    if not vertices_only:
        ffn = os.path.join(space, 'geometry', 'faces', f'{lr}h.npy')
        faces = core_dataset.get(ffn, on_missing='raise')
        if which == 'faces':
            return faces

    if which in ['sphere', 'sphere.reg']:
        fn = os.path.join(space, 'geometry', 'sphere.reg', f'{lr}h.npy')
    else:
        fn = os.path.join(space, 'geometry', which, f'{lr}h', f'{group}_{avg_type}.npy')
    coords = core_dataset.get(fn, on_missing='raise')
    if vertices_only:
        return coords

    return coords, faces


def get_distances(
    lr, source, target=None, mask=None, source_mask=None, target_mask=None, **kwargs
):
    """Distances between vertices.

    Parameters
    ----------
    lr : str
        Hemisphere, either 'l' or 'r'.
    source : str
        Source space.
    target : str, default=None
        Target space. If None, assuming it's the same as ``source``.
    mask : str, default=None
        Mask to apply to the distance matrix. If None or False, no mask is
        applied. If True, the group-based cortical mask is used.
        If a boolean array, it is used as the mask.
        ``source_mask`` and ``target_mask`` can be set separately, and they
        take precedence over ``mask``.
    source_mask : str, default=None
        Mask to apply to source space (rows) of the distance matrix.
    target_mask : str, default=None
        Mask to apply to target space (columns) of the distance matrix.

    Returns
    -------
    M : ndarray
        Distance matrix. The shape is (n_source_vertices, n_target_vertices).
    """
    group = kwargs.get('group', 'on1031')
    avg_type = kwargs.get('avg_type', 'trimmed')
    dist_type = kwargs.get('dist_type', 'dijkstra')

    if target is None:
        target = source
    if source_mask is None:
        source_mask = mask
    if target_mask is None:
        target_mask = mask

    fn = os.path.join(
        source,
        'distances',
        f'to_{target}',
        f'{lr}h',
        f'{group}_{avg_type}',
        f'{dist_type}.npy',
    )

    assert target == source
    d = core_dataset.get(fn, on_missing='raise')
    ico = int(source.split('-ico')[1])
    nv = ico**2 * 10 + 2
    mat = np.zeros((nv, nv), dtype=d.dtype)
    idx1, idx2 = np.triu_indices(nv, 1)
    mat[idx1, idx2] = d
    mat = np.maximum(mat, mat.T)
    del d

    if source_mask is not None:
        if isinstance(source_mask, np.ndarray):
            mask1 = source_mask
        else:
            mask1 = get_mask(lr, source, **kwargs)
    if target_mask is not None:
        if isinstance(target_mask, np.ndarray):
            mask2 = target_mask
        else:
            mask2 = get_mask(lr, target, **kwargs)

    if source_mask is not None and target_mask is not None:
        M = mat[np.ix_(mask1, mask2)]
    elif source_mask is not None:
        M = mat[mask1, :]
    elif target_mask is not None:
        M = mat[:, mask2]
    else:
        M = mat

    return M


def smooth(lr, fwhm, space='onavg-ico32', mask=None, keep_sum=False):
    """Get a smoothing matrix.

    Parameters
    ----------
    lr : str
        Hemisphere, either 'l' or 'r'.
    fwhm : float
        Full-width at half-maximum of the Gaussian kernel.
    space : str, default='onavg-ico32'
        Surface space where the data is in.
    mask : ndarray or bool or None, default=None
        Mask to apply to the smoothing matrix. If None or False, no mask is
        applied. If True, the standard cortical mask is used. If an ndarray,
        it is used as the mask.
    keep_sum : bool, default=False
        If True, keep the sum of the data. Useful for transforming area,
        volume, etc., where the total area/volume is preserved when
        ``keep_sum=True``.

    Returns
    -------
    M : sparse matrix
        Smoothing matrix. Can be applied to data matrix ``X`` as ``X @ M``.
    """
    d = get_distances(lr, space, space, mask=mask)
    s2 = fwhm / (4.0 * np.log(2))
    weights = np.exp(-(d**2) / s2)
    mat = sparse.csr_matrix(weights)

    if keep_sum:
        with np.errstate(divide='ignore'):
            d = np.nan_to_num(np.reciprocal(mat.sum(axis=1).A.ravel()))
        M = sparse.diags(d) @ mat
    else:
        with np.errstate(divide='ignore'):
            d = np.nan_to_num(np.reciprocal(mat.sum(axis=0).A.ravel()))
        M = mat @ sparse.diags(d)

    return M


def get_mapping(
    lr,
    source,
    target,
    mask=None,
    nn=False,
    keep_sum=False,
    source_mask=None,
    target_mask=None,
    **kwargs,
):
    """Get mapping (transform) from one space to another.

    Parameters
    ----------
    lr : str
        Hemisphere, either 'l' or 'r'.
    source : str
        Source space, the space where the data is currently in.
    target : str
        Target space, the space where the data will be transformed into.
    mask : bool or boolean array or None
        Mask to apply to the mapping. If None or False, no mask is applied.
        If True, the group-based cortical mask is used.
        If a boolean array, it is used as the mask.
        ``source_mask`` and ``target_mask`` can be set separately, and they
        take precedence over ``mask``.
    nn : bool
        If True, use nearest neighbor interpolation.
        If False, use overlap-area-based interpolation.
    keep_sum : bool
        If True, keep the sum of the data. Useful for transforming area,
        volume, etc., where the total area/volume is preserved when
        ``keep_sum=True``.
    source_mask : bool or boolean array or None
        Mask to apply to the source space. Similar to ``mask``.
    target_mask : bool or boolean array or None
        Mask to apply to the target space. Similar to ``mask``.

    Returns
    -------
    M : sparse matrix
        Mapping matrix. Can be applied to data in the source space to
        transform it into the target space. For example, if ``X`` is a
        data matrix in the source space, then ``X @ M`` is the data matrix
        in the target space.
    """
    species = (
        'macaque' if source.startswith('mkavg-') or source == 'MEBRAIN' else 'human'
    )
    if species == 'human':
        group, avg_type = 'on1031', 'trimmed'
    elif species == 'macaque':
        group, avg_type = 'mk12', 'average'
    else:
        raise ValueError(f'Unknown species: {species}')

    group = kwargs.get('group', group)
    resample = kwargs.get('resample', 'overlap-8div')
    avg_type = kwargs.get('avg_type', avg_type)

    if source_mask is None:
        source_mask = mask
    if target_mask is None:
        target_mask = mask

    if source == target:
        ico = int(source.split('-ico')[1])
        nv = ico**2 * 10 + 2
        mat = sparse.diags(np.ones((nv,))).tocsr()
        # print(mat.shape, mat.data.shape, type(mat))
    else:
        fn1 = os.path.join(
            source,
            'mapping',
            f'to_{target}',
            f'{lr}h',
            f'{group}_{avg_type}',
            f'{resample}.npz',
        )
        fn2 = os.path.join(
            target,
            'mapping',
            f'to_{source}',
            f'{lr}h',
            f'{group}_{avg_type}',
            f'{resample}.npz',
        )
        mat1 = core_dataset.get(fn1, on_missing='ignore')
        mat2 = core_dataset.get(fn2, on_missing='ignore')
        assert mat1 is not None or mat2 is not None, f'Neither {fn1} nor {fn2} exists.'
        mat = mat1 if mat1 is not None else mat2.T

    if source_mask is not None and source_mask is not False:
        if isinstance(source_mask, np.ndarray):
            mask1 = source_mask
        else:
            mask1 = get_mask(lr, source, **kwargs)
    if target_mask is not None and target_mask is not False:
        if isinstance(target_mask, np.ndarray):
            mask2 = target_mask
        else:
            mask2 = get_mask(lr, target, **kwargs)

    if source_mask is not None and target_mask is not None:
        mat = mat[np.ix_(mask1, mask2)]
    elif source_mask is not None:
        mat = mat[mask1, :]
    elif target_mask is not None:
        mat = mat[:, mask2]
    else:
        pass

    if nn:
        idx = mat.argmax(axis=0).A.ravel()
        m, n = mat.shape
        M = sparse.csr_matrix((np.ones((n,)), (idx, np.arange(n))), (m, n))
        return M

    if keep_sum:
        with np.errstate(divide='ignore'):
            d = np.nan_to_num(np.reciprocal(mat.sum(axis=1).A.ravel()))
        M = sparse.diags(d) @ mat
    else:
        with np.errstate(divide='ignore'):
            d = np.nan_to_num(np.reciprocal(mat.sum(axis=0).A.ravel()))
        M = mat @ sparse.diags(d)

    return M



import os
import numpy as np
import neuroboros as nb

from .io import core_dataset


def load_npz(npz_fn):
    npz = np.load(npz_fn)
    sls = np.array_split(npz['concatenated_searchlights'], npz['boundaries'])
    dists = np.array_split(npz['concatenated_distances'], npz['boundaries'])
    return sls, dists


def load_searchlights(lr, radius, space, center_space=None, **kwargs):
    species = 'macaque' if space.startswith('mkavg-') else 'human'
    if species == 'human':
        group, dist_type, avg_type = 'on1031', 'dijkstra', 'trimmed'
    elif species == 'macaque':
        group, dist_type, avg_type = 'mk12', 'geodesic', 'average'
    else:
        raise ValueError(f'Unknown species: {species}')
    group = kwargs.get('group', group)
    dist_type = kwargs.get('dist_type', dist_type)
    avg_type = kwargs.get('avg_type', avg_type)
    if center_space is None:
        center_space = space
    npz_fn = os.path.join(
        space,
        'searchlights',
        f'{center_space}_center',
        f'{lr}h',
        f'{group}_{avg_type}',
        f'{dist_type}_{radius}mm.npz',
    )
    sls, dists = core_dataset.get(npz_fn, load_func=load_npz, on_missing='raise')
    return sls, dists


def convert_searchlights(sls, dists, radius, mask, center_mask):
    """
    Convert larger-radius/unmasked searchlights to smaller-radius/masked.

    Parameters
    ----------
    sls : list of ndarray
        Each entry is an ndarray of integers, which are the indices of the
        vertices in a searchlight.
    dists : list of ndarray
        Each entry is an ndarray of float numbers, which are the distances
        between vertices in a searchlight and the center of the
        searchlight. The order of vertices are the same as ``sls``.
    radius : int or float
        The searchlight radius in mm.
    mask : ndarray
        A boolean ndarray indicating whether a vertex belongs to the cortex
        (True) or not (False).
    center_mask : ndarray
        A boolean ndarray indicating whether a searchlight center belongs to
        the cortex (True) or not (False). Mainly used in sparse searchlights
        where the number of searchlights is smaller than the number of
        vertices.

    Returns
    -------
    sls_new : list of ndarray
        The new list of searchlight indices after conversion.
    dists_new : list of ndarray
        The new list of distances to searchlight center after conversion.
    """
    if mask is not False:
        cortical_indices = np.where(mask)[0]
        mapping = np.cumsum(mask) - 1

    sls_new, dists_new = [], []
    for i, (sl, d) in enumerate(zip(sls, dists)):
        if center_mask is False or center_mask[i]:
            m = d <= radius
            if mask is not False:
                m = np.logical_and(np.isin(sl, cortical_indices), m)
                sls_new.append(mapping[sl[m]])
            else:
                sls_new.append(sl[m])
            dists_new.append(d[m])

    return sls_new, dists_new


def get_searchlights(
    lr,
    radius,
    space='onavg-ico32',
    center_space=None,
    mask=True,
    center_mask=None,
    return_dists=False,
    **kwargs,
):
    radius_ = 20

    if center_mask is None:
        center_mask = mask

    if mask and not isinstance(mask, np.ndarray):
        mask = nb.mask(lr, space, **kwargs)

    if center_space is None:
        center_space = space
        center_mask = mask
    else:
        if center_mask is True:
            center_mask = nb.mask(lr, center_space, **kwargs)

    sls, dists = load_searchlights(
        lr, radius_, space, center_space=center_space, **kwargs
    )
    sls, dists = convert_searchlights(sls, dists, radius, mask, center_mask)

    if return_dists:
        return sls, dists
    return sls



"""
===================================================
Linear algebra utilities (:mod:`neuroboros.linalg`)
===================================================

.. currentmodule:: neuroboros.linalg

.. autosummary::
    :toctree:

    safe_svd - Singular value decomposition without occasional LinAlgError crashes.
    safe_polar - Polar decomposition without occasional LinAlgError crashes.
    gram_pca - Principal component analysis based on the Gram matrix.

"""
import numpy as np
from scipy.linalg import LinAlgError, eigh, polar, svd


def safe_svd(X, remove_mean=True):
    """
    Singular value decomposition without occasional LinAlgError crashes.

    The default ``lapack_driver`` of ``scipy.linalg.svd`` is ``'gesdd'``,
    which occasionally crashes even if the input matrix is not singular.
    This function automatically handles the ``LinAlgError`` when it's
    raised and switches to the ``'gesvd'`` driver in this case.

    The input matrix ``X`` is factorized as ``U @ np.diag(s) @ Vt``.

    Parameters
    ----------
    X : ndarray of shape (M, N)
        The matrix to be decomposed in NumPy array format.
    remove_mean : bool, default=True
        Whether to subtract the mean of each column before the actual SVD
        (True) or not (False). Setting ``remove_mean=True`` is helpful when
        the SVD is used to perform PCA.

    Returns
    -------
    U : ndarray of shape (M, K)
        Unitary matrix.
    s : ndarray of shape (K,)
        The singular values.
    Vt : ndarray of shape (K, N)
        Unitary matrix.
    """
    if remove_mean:
        mean = X.mean(axis=0, keepdims=True)
        if not np.allclose(mean, 0, atol=1e-10):
            X = X - mean

    try:
        U, s, Vt = svd(X, full_matrices=False)
    except LinAlgError:
        U, s, Vt = svd(X, full_matrices=False, lapack_driver='gesvd')

    return U, s, Vt


def safe_polar(a, side='left'):
    """
    Polar decomposition without occasional LinAlgError crashes.

    The default ``lapack_driver`` of ``scipy.linalg.polar`` is ``'gesdd'``,
    which occasionally crashes even if the input matrix is not singular.
    This function automatically handles the ``LinAlgError`` when it's
    raised and switches to the ``'gesvd'`` driver in this case.

    The input matrix ``a`` is factorized as ``u @ p`` (or ``p @ u`` when
    ``side='right'``).

    Parameters
    ----------
    a : ndarray of shape (M, N)
        The matrix to be decomposed in NumPy array format.
    side : {'left', 'right'}, default='left'
        Whether to return ``u @ p`` (``side='left'``) or ``p @ u``
        (``side='right'``).

    Returns
    -------
    u : ndarray of shape (M, M) or (N, N)
        Unitary matrix (rotation and reflection).
    p : ndarray of shape (M, N) or (N, M)
        Hermitian positive semi-definite matrix (scaling and shearing).

    """

    try:
        u, p = polar(a, side=side)
    except Exception as e:
        w, s, vh = safe_svd(a)
        u = w.dot(vh)
        if side == 'right':
            # a = up
            p = (vh.T.conj() * s).dot(vh)
        else:
            # a = pu
            p = (w * s).dot(w.T.conj())
    return u, p


def gram_pca(gram, tol=1e-7):
    """
    Principal component analysis (PCA) based on the Gram matrix.

    Parameters
    ----------
    gram : ndarray of shape (N, N)
        The Gram matrix to be decomposed in NumPy array format.
    tol : float, default=1e-7
        Tolerance for the eigenvalues to be considered positive.

    Returns
    -------
    PCs : ndarray of shape (N, N - 1)
        The principal components (PCs) derived from the Gram matrix.
    """

    w, v = eigh(gram, lower=False)
    assert np.all(w > -tol)
    w[w < 0] = 0
    U = v[:, ::-1][:, :-1]
    s = np.sqrt(w[::-1][:-1])
    PCs = U * s[np.newaxis]
    return PCs



import numpy as np
from scipy.stats import zscore

import neuroboros as nb


def inverse_spearman_brown(rs, n):
    rs1 = 1.0 / (1.0 + n * (1.0 - rs) / rs)
    return rs1


class TestSpearmanBrown:
    def test_spearman_brown(self):
        rng = np.random.default_rng(0)
        for i in range(10):
            r1 = rng.random()
            n = rng.integers(2, 100)
            rn = nb.stats.spearman_brown(r1, n)
            assert rn > r1
            assert rn <= 1
            inv = inverse_spearman_brown(rn, n)
            np.testing.assert_allclose(inv, r1)

    def test_spearman_brown_ndarray(self):
        rng = np.random.default_rng(0)
        for i in range(10):
            r1 = rng.random((4, 5))
            n = rng.integers(2, 100, size=(4, 5))
            rn = nb.stats.spearman_brown(r1, n)
            np.testing.assert_array_less(r1, rn)
            np.testing.assert_array_less(rn, 1.0)
            inv = inverse_spearman_brown(rn, n)
            np.testing.assert_allclose(inv, r1)

    def test_spearman_brown_inv(self):
        rng = np.random.default_rng(0)
        for i in range(100):
            rn = rng.random()
            n = rng.integers(2, 100)
            inv = inverse_spearman_brown(rn, n)
            r1 = nb.stats.spearman_brown(rn, 1.0 / n)
            np.testing.assert_allclose(inv, r1)


class TestSpearmanBrownInv:
    def test_spearman_brown_inv(self):
        rng = np.random.default_rng(0)
        for i in range(100):
            r1 = rng.random()
            rn = rng.random()
            if r1 == 0 or rn == 0:
                continue
            n = nb.stats.spearman_brown_inv(r1, rn)
            r2 = nb.stats.spearman_brown(r1, n)
            np.testing.assert_allclose(r2, rn)


class TestAlpha:
    def test_alpha(self):
        rng = np.random.default_rng(0)
        n = 100
        for i in range(10):
            X = rng.standard_normal((2, n))
            X = zscore(X, axis=1)
            alpha = nb.stats.cronbach_alpha(X)
            r1 = np.mean(X[0] * X[1])
            r2 = nb.stats.spearman_brown(r1, 2)
            np.testing.assert_allclose(alpha, r2)

    def test_alpha_ndarray(self):
        rng = np.random.default_rng(0)
        n = 100
        m = 5
        for i in range(10):
            X = rng.standard_normal((2, m, n))
            alphas1 = nb.stats.cronbach_alpha(X, var_axis=2)
            alphas2 = np.array([nb.stats.cronbach_alpha(X[:, i]) for i in range(m)])
            np.testing.assert_array_equal(alphas1, alphas2)

    def test_alpha_ci(self):
        rng = np.random.default_rng(0)
        n = 100
        m = 5
        for i in range(10):
            X = rng.standard_normal((2, m, n))
            alphas1, cis1 = nb.stats.cronbach_alpha(X, var_axis=2, ci=0.95)
            returns = [
                nb.stats.cronbach_alpha(X[:, i], ci=[0.025, 0.975]) for i in range(m)
            ]
            alphas2 = np.stack([_[0] for _ in returns])
            cis2 = np.stack([_[1] for _ in returns], axis=1)
            np.testing.assert_array_equal(alphas1, alphas2)
            np.testing.assert_array_equal(cis1, cis2)
        np.testing.assert_array_less(cis1[:, 0], cis1[:, 1])



import numpy as np
import scipy.sparse as sparse

from .dijkstra import dijkstra
from .properties import compute_neighbor_distances, compute_vertex_areas
from .union import compute_union_coords


def compute_vertex_nn(nv, indices2, neighbors, neighbor_distances):
    nn = np.zeros((nv,), dtype=int)
    nnd = np.full((nv,), np.inf)
    for src in indices2:
        nbrs, dists = dijkstra(src, nv, neighbors, neighbor_distances, max_dist=3.0)
        for nbr, d in zip(nbrs, dists):
            if nnd[nbr] > d:
                nnd[nbr] = d
                nn[nbr] = src

    for max_dist in [4.0, 5.0, 6.0, 8.0, 10.0, np.inf]:
        for src in np.where(np.isinf(nnd))[0]:
            nbrs, dists = dijkstra(
                src, nv, neighbors, neighbor_distances, max_dist=max_dist
            )
            idx = np.where(np.isin(nbrs, indices2))[0]
            if len(idx):
                nnd[src] = dists[idx[0]]
                nn[src] = nbrs[idx[0]]

    mapping = np.zeros((nv,), dtype=int)
    mapping[indices2] = np.arange(len(indices2))
    nn = mapping[nn]

    return nn


def areal(sphere, coords1, anat_coords, coords2=None):
    if coords2 is not None:
        coords12, indices1in12, indices2in12 = compute_union_coords(coords1, coords2)
        combined_sphere, indices0, indices12 = sphere.union(coords12)
        indices1, indices2 = indices12[indices1in12], indices12[indices2in12]
    else:
        combined_sphere, indices0, indices1 = sphere.union(coords1)
        indices1, indices2 = indices0, indices1

    xform = sphere.barycentric(combined_sphere.coords)
    combined_coords = xform.T @ anat_coords
    assert combined_coords.shape == combined_sphere.coords.shape
    nv, neighbors = combined_sphere.nv, combined_sphere.neighbors
    neighbor_distances = compute_neighbor_distances(combined_coords, neighbors)

    nn1 = compute_vertex_nn(nv, indices1, neighbors, neighbor_distances)
    nn2 = compute_vertex_nn(nv, indices2, neighbors, neighbor_distances)
    vertex_areas = compute_vertex_areas(combined_coords, combined_sphere.faces)
    mat = sparse.lil_matrix((indices1.shape[0], indices2.shape[0]))

    for n1, n2, va in zip(nn1, nn2, vertex_areas):
        mat[n1, n2] += va
    mat = mat.tocsr()

    return mat



import numpy as np
from scipy.stats import f, norm, rankdata, zscore


def cronbach_alpha_ci(alpha, n, m, ci):
    """Confidence interval for Cronbach's alpha.

    Parameters
    ----------
    alpha : float
        Cronbach's alpha.
    n : int
        Number of repetitions, e.g., items, parallel tests, repeated measures
        of brain response patterns.
    m : int
        Number of individuals, e.g., test-takers, voxels, vertices.
    ci : float or tuple of float
        Confidence level. If float, it should be between 0 and 1, e.g., 0.95
        for 95% confidence interval. If tuple, it should be a pair of floats
        between 0 and 1, e.g., (0.025, 0.975) for 95% confidence interval.

    Returns
    -------
    ci : tuple of float
        Confidence interval.

    Notes
    -----
    Feldt, L. S., Woodruff, D. J., & Salih, F. A. (1987). Statistical Inference for Coefficient Alpha. Applied Psychological Measurement, 11(1), 93–103. https://doi.org/10.1177/014662168701100107

    """
    if isinstance(ci, (tuple, list)):
        u, l = ci
    else:
        assert ci >= 0 and ci <= 1
        u = (1 - ci) * 0.5
        l = 1 - u
    ratios = [f.ppf(_, m - 1, (n - 1) * (m - 1)) for _ in [l, u]]
    ci = tuple(1 - (1 - alpha) * _ for _ in ratios)
    return ci


def cronbach_alpha(X, rep_axis=0, var_axis=1, ci=None, squeeze=True):
    """Cronbach's alpha.

    Parameters
    ----------
    X : ndarray
        The data.
    rep_axis : int, default=0
        The axis along which the repetitions are stacked, e.g., items,
        parallel tests, repeated measures of brain response patterns.
    var_axis : int, default=1
        The axis along which individuals differ, e.g., test-takers, voxels,
        vertices.
    ci : float or tuple of float or None, default=None
        Confidence level. If float, it should be between 0 and 1, e.g., 0.95
        for 95% confidence interval. If tuple, it should be a pair of floats
        between 0 and 1, e.g., (0.025, 0.975) for 95% confidence interval.
        If None, no confidence interval is computed or returned.
    squeeze : bool, default=True
        Whether to squeeze the returned ndarray for ``rep_axis`` and
        ``var_axis``. If False, the returned ndarray has the same number of
        dimensions as ``X``.

    Returns
    -------
    alpha : float or ndarray
        Cronbach's alpha. If ``X`` is 2D, a float is returned. Otherwise, an
        ndarray is returned.
    ci : ndarray
        Confidence interval. Returned only if ``ci`` is not None.
    """
    v1 = X.var(axis=var_axis, ddof=1, keepdims=True).sum(axis=rep_axis, keepdims=True)
    v2 = X.sum(axis=rep_axis, keepdims=True).var(axis=var_axis, ddof=1, keepdims=True)
    if squeeze:
        v1 = np.squeeze(v1, axis=(rep_axis, var_axis))
        v2 = np.squeeze(v2, axis=(rep_axis, var_axis))
    n = X.shape[rep_axis]
    alpha = (n / (n - 1.0)) * (1.0 - v1 / v2)
    if ci is not None:
        m = X.shape[var_axis]
        ci = cronbach_alpha_ci(alpha, n, m, ci)
        ci = np.stack(ci, axis=0)
        return alpha, ci
    return alpha


def noise_ceilings(X, rep_axis=0, var_axis=1, return_alpha=False):
    """Noise ceilings.

    Parameters
    ----------
    X : ndarray
        The data.
    rep_axis : int
        The axis along which the repetitions are stacked, e.g., items,
        parallel tests, repeated measures of brain response patterns.
    var_axis : int
        The axis along which individuals differ, e.g., test-takers, voxels,
        vertices.
    return_alpha : bool, default=False
        Whether to return Cronbach's alpha coefficients based on the n - 1
        individuals during the lower bound estimates.

    Returns
    -------
    ceilings : ndarray
        The noise ceilings. The first axis has two elements, the first element
        is the lower bound, and the second element is the upper bound. If
        ``return_alpha`` is True, the first axis has three elements, the third
        element is the Cronbach's alpha coefficient.
    """
    n = X.shape[rep_axis]
    if return_alpha:
        alpha = [
            cronbach_alpha(
                np.delete(X, i, axis=rep_axis), rep_axis, var_axis, squeeze=False
            )
            for i in range(n)
        ]
        alpha = np.concatenate(alpha, axis=rep_axis)
        alpha = np.squeeze(alpha, axis=var_axis)

    s = X.sum(axis=rep_axis, keepdims=True)
    d = zscore(s - X, axis=var_axis)
    X = zscore(X, axis=var_axis)
    lower = np.mean(X * d, axis=var_axis)
    s = zscore(s, axis=var_axis)
    upper = np.mean(X * s, axis=var_axis)
    if return_alpha:
        ceilings = np.stack([lower, upper, alpha], axis=0)
    else:
        ceilings = np.stack([lower, upper], axis=0)
    return ceilings


def spearman_brown(r, n):
    """Spearman-Brown prediction formula.

    Parameters
    ----------
    r : float
        The reliability coefficient.
    n : int or float
        The length of the new "test" relative to the current one. For example,
        if the current test has 10 items and the new test has 20 items, then
        ``n`` is 2. If the current test has 20 items and the new test has 10
        items, then ``n`` is 0.5.

    Returns
    -------
    r_new : float
        The predicted reliability coefficient.
    """
    r_new = (n * r) / (1.0 + (n - 1.0) * r)
    return r_new


def spearman_brown_inv(r, r_new):
    """Predicting required test length for a given reliability.

    Parameters
    ----------
    r : float
        The current reliability coefficient.
    r_new : float
        The desired reliability coefficient.

    Returns
    -------
    n : float
        The required test length relative to the current one.
    """
    prod = r * r_new
    n = (r_new - prod) / (r - prod)
    return n


def normalize(d, clip=(0.0005, 0.9995), keep_stats=True):
    """Normalize the data to normal distribution.

    Parameters
    ----------
    d : ndarray
        The data to be normalized. The function assumes that ``d`` is 1D.
    clip : float or tuple of float, default=(0.0005, 0.9995)
        The lower and upper bounds for clipping the percentiles. By default,
        the function clips percentiles above 99.95% and below 0.05%. If a float
        is provided, it is used as the percentage of clipping, e.g., 0.01 for
        clipping percentiles above 99.5% and below 0.5%.
    keep_stats : bool, default=True
        Whether to keep the mean and standard deviation of the input data. If
        False, the returned data is normalized to standard normal distribution
        (zero mean and unit variance).

    Returns
    -------
    d_new : ndarray
        The normalized data.

    """
    rank = rankdata(d)
    pct = (rank - 0.5) / len(rank)

    if isinstance(clip, float):
        clip = (clip * 0.5, 1 - clip * 0.5)
    pct = np.clip(pct, *clip)

    if keep_stats:
        d_new = norm.ppf(pct, loc=d.mean(), scale=d.std())
    else:
        d_new = norm.ppf(pct, loc=0, scale=1)
    return d_new



"""
================================================
General Linear Model (:mod:`neuroboros.glm`)
================================================

.. currentmodule:: neuroboros.glm

.. autosummary::
    :toctree:

"""

import numpy as np


def glm(dm, design, nuisance=None, contrasts=None, return_r2=False):
    """General linear model.

    Parameters
    ----------
    dm : array, shape (n_samples, n_features)
        Data matrix.
    design : array, shape (n_samples, n_design_regressors)
        Experimental design regressors.
    nuisance : array, shape (n_samples, n_nuisance_regressors), optional
        Nuisance regressors.
    contrasts : array, shape (n_contrasts, n_design_regressors), optional
        Contrasts. If None, identity matrix is used, which models the response
        to each design regressor separately.

    Returns
    -------
    betas : array, shape (n_contrasts, n_features)
        Estimated beta coefficients.
    ts : array, shape (n_contrasts, n_features)
        Estimated t-statistics.
    R2s : array, shape (n_features,)
        Adjusted R-squared values. Only returned if `return_r2` is True.
    """
    if nuisance is not None:
        regressors = np.concatenate([design, nuisance], axis=1)
    else:
        regressors = design
    nr = regressors.shape[1]

    if contrasts is None:
        contrasts = np.eye(design.shape[1])

    beta = np.linalg.lstsq(regressors, dm, rcond=None)[0]
    diff = dm - regressors @ beta
    ss_res = np.sum(diff**2, axis=0)
    dof_res = dm.shape[0] - nr
    sigma = np.sqrt(ss_res / dof_res)
    cov = regressors.T @ regressors
    inv = np.linalg.inv(cov)

    # TODO all contrasts at once based on
    # https://torwager.github.io/ComputationalFoundations/glm2_param_inference.html
    betas, ts = [], []
    for contrast in contrasts:
        R = np.concatenate([contrast, [0] * (nr - len(contrast))]).reshape(1, -1)
        mid = R @ inv @ R.T
        assert mid.shape == (1, 1)
        mid = 1.0 / mid[0, 0]
        ratio = np.sqrt(mid) / sigma
        R_beta = (R @ beta).ravel()
        t = R_beta * ratio
        betas.append(R_beta)
        ts.append(t)

    betas, ts = np.array(betas), np.array(ts)

    if return_r2:
        if nuisance is None:
            ss_all = np.sum(dm**2, axis=0)
            R2s = 1 - ss_res * dm.shape[0] / (ss_all * dof_res)
        else:
            beta2 = np.linalg.lstsq(nuisance, dm, rcond=None)[0]
            ss_all = np.sum((dm - nuisance @ beta2) ** 2, axis=0)
            dof_all = dm.shape[0] - nuisance.shape[1]
            R2s = 1 - ss_res * dof_all / (ss_all * dof_res)
        return betas, ts, R2s

    return betas, ts



import numpy as np
from scipy.stats import ttest_ind

from neuroboros import glm


def test_glm():
    rng = np.random.default_rng()
    dm = rng.standard_normal((1000, 500))
    nuisance = rng.standard_normal((1000, 10))
    design = rng.standard_normal((1000, 10))
    betas, ts, R2s = glm(dm, design, nuisance, return_r2=True)

    assert betas.shape == (10, 500)
    assert ts.shape == (10, 500)
    assert R2s.shape == (500,)

    assert np.allclose(np.mean(R2s), 0, atol=0.001)
    assert np.allclose(np.mean(betas), 0, atol=0.001)


def test_t_value():
    rng = np.random.default_rng()
    dm = rng.standard_normal((100, 1))
    nuisance = None
    design = rng.choice([0, 1], (100, 1))
    design = np.concatenate([design, np.ones_like(design)], axis=1)
    ts = glm(dm, design, nuisance)[1]
    t = ttest_ind(dm[design[:, 0] == 1], dm[design[:, 0] == 0], equal_var=True)
    print(ts[0], t.statistic)
    assert np.allclose(ts[0], t.statistic, atol=0.001)



import os
import warnings

import requests

try:
    import datalad.api as dl
except ImportError:
    dl = None

from .utils import load

DATA_ROOT = os.environ.get(
    'NEUROBOROS_DATA_DIR', os.path.join(os.path.expanduser('~'), '.neuroboros-data')
)


class DatasetManager:
    def __init__(self, name, root=None, source=None, kind=None):
        self.name = name
        self.root = os.path.join(DATA_ROOT, name) if root is None else root
        self.source = source
        self.kind = kind

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, kind):
        if kind is None:
            if self.source is None:
                kind = 'local'
            elif os.path.exists(os.path.join(self.root, '.git')) and dl is not None:
                kind = 'datalad'
            else:
                kind = 'alternative'
        else:
            assert kind in ['local', 'datalad', 'alternative']
        self._kind = kind

        if self._kind == 'local':
            assert os.path.exists(
                self.root
            ), f"Local dataset {self.name} does not exist at {self.root}."
        elif self._kind == 'datalad':
            if os.path.exists(self.root):
                self.dset = dl.Dataset(self.root)
                assert self.dset.repo is not None
                urls = [
                    self.dset.repo.get_remote_url(remote)
                    for remote in self.dset.repo.get_remotes()
                ]
                if self.source not in urls:
                    warnings.warn(
                        f"DataLad dataset {self.name} exists at {self.root}, but "
                        f"does not have source {self.source} as a sibling."
                    )
            else:
                self.dset = dl.clone(self.source, self.root)
        elif self._kind == 'alternative':
            if self.source.startswith('https://gin.g-node.org/'):
                self.url_base = self.source + '/raw/master/'
            else:
                raise NotImplementedError(
                    "Only supports downloading data from GIN without DataLad."
                )

        self.download = getattr(self, f'_download_{kind}')

    def get(self, fn, load_func=None, on_missing='warn'):
        if isinstance(fn, (tuple, list)):
            local_fn = os.path.join(self.root, *fn)
        else:
            local_fn = os.path.join(self.root, fn)

        if not os.path.exists(local_fn):
            self.download(fn, local_fn, on_missing=on_missing)
        if not os.path.exists(local_fn):
            return None

        if load_func is None:
            return load(local_fn)
        return load_func(local_fn)

    def _download_local(self, fn, local_fn, on_missing='warn'):
        if on_missing == 'raise':
            raise RuntimeError(f"File {local_fn} not found.")
        elif on_missing == 'warn':
            warnings.warn(f"File {local_fn} not found.")
            return None
        elif on_missing == 'ignore':
            return None
        else:
            raise ValueError(f"Invalid value for `on_missing`: {on_missing}")

    def _download_datalad(self, fn, local_fn, on_missing='warn'):
        if not os.path.lexists(local_fn):
            if on_missing == 'raise':
                raise RuntimeError(f"File {local_fn} not found in DataLad repo.")
            elif on_missing == 'warn':
                warnings.warn(f"File {local_fn} not found in DataLad repo.")
                return None
            elif on_missing == 'ignore':
                return None
            else:
                raise ValueError(f"Invalid value for `on_missing`: {on_missing}")

        result = self.dset.get('/'.join(fn))[0]
        if result['status'] not in ['ok', 'notneeded']:
            raise RuntimeError(
                f"datalad `get` status is {result['status']}, likely due "
                "to problems downloading the file."
            )
        assert os.path.normpath(result['path']) == os.path.normpath(local_fn)

    def _download_alternative(self, fn, local_fn, on_missing='warn'):
        if isinstance(fn, (tuple, list)):
            url = self.url_base + '/'.join(fn)
        else:
            url = self.url_base + fn.replace('\\', '/')

        try:
            r = requests.get(url)
        except requests.exceptions.RequestException as e:
            raise RuntimeError(f"Error downloading {url}: {e}")
        else:
            if r.status_code == 404:
                if on_missing == 'raise':
                    raise RuntimeError(f"File {url} not found.")
                elif on_missing == 'warn':
                    warnings.warn(f"File {url} not found.")
                    return None
                elif on_missing == 'ignore':
                    return None
                else:
                    raise ValueError(f"Invalid value for `on_missing`: {on_missing}")
            if r.status_code != 200:
                raise RuntimeError(f"Error downloading {url}: {r.status_code}")
        os.makedirs(os.path.dirname(local_fn), exist_ok=True)
        with open(local_fn, 'wb') as f:
            f.write(r.content)


# DefaultDataset = DatasetManager
core_dataset = DatasetManager('core', source='https://gin.g-node.org/neuroboros/core')



import numpy as np
from scipy.spatial.distance import cdist


def classification(
    U, Uhat, size=None, npc=None, metric='correlation', reduce_func=np.mean
):
    """Movie time point/segment classification accuracy.

    Parameters
    ----------
    U : ndarray, shape (nt, nv)
        The measured response patterns, often the columns are normalized PCs of
        the measured data rather than the original voxels/vertices.
    Uhat : ndarray, shape (nt, nv)
        The predicted response patterns.
    size : int or None, default=None
        The number of time points in each segment. If None, each time point is
        treated as a segment.
    npc : int or None, default=None
        The number of PCs used for classification. If None, all PCs are used.
        If not None, assuming that ``U`` and ``Uhat`` are PCs, the first
        ``npc`` PCs are used for classification.
    metric : str, default='correlation'
        The metric used for calculating the distance between two segments.
        See ``scipy.spatial.distance.cdist`` for the available metrics.
    reduce_func : callable or None, default=np.mean
        The function used for reducing the classification accuracy across
        segments. If None, the raw classification accuracy is returned, where
        the length of the returned array is equal to the number of segments.

    Returns
    -------
    acc : float or ndarray
        The classification accuracy. If ``reduce_func`` is None, the returned
        value is an ndarray of shape (nc,), where nc is the number of segments.
        Otherwise, the returned value is a float.
    """
    U, Uhat = U[:, :npc], Uhat[:, :npc]
    nt, nv = U.shape
    if size is None:
        nc = nt
        Y, Yhat = U, Uhat
    else:
        nc = nt // size
        Y = U[: nc * size].reshape(nc, size, nv).reshape(nc, size * nv)
        Yhat = Uhat[: nc * size].reshape(nc, size, nv).reshape(nc, size * nv)
    d = cdist(Y, Yhat, metric=metric)
    arng = np.arange(nc)
    acc = np.argmin(d, axis=1) == arng
    if reduce_func is not None:
        acc = reduce_func(acc)
    return acc



import io
import os
from warnings import warn

import numpy as np
from matplotlib import cm, colors

try:
    from PIL import Image as PIL_Image

    PIL_ok = True
except ImportError as e:
    PIL_ok = False

from .io import core_dataset
from .plot2d import Image
from .spaces import get_mapping, get_mask
from .utils import save

GUESS_SEPARATE = {
    (10242, 10242): ('mkavg-ico32', False),
    (100988, 100974): ('MEBRAIN', False),
}

GUESS_COMBINED = {
    20484: ('mkavg-ico32', False, [10242]),
    201962: ('MEBRAIN', False, [100988]),
}

PLOT_MAPPING = {}


def unmask_and_upsample(values, space, mask, nn=True):
    if space is None and mask is None:
        if isinstance(values, np.ndarray):
            ret = GUESS_COMBINED[values.shape[0]]
            if len(ret) == 4:
                space, mask, boundary, flavor = ret
                mask_kwargs = {'flavor': flavor, 'legacy': True}
            else:
                space, mask, boundary = ret
                mask_kwargs = {}
        elif isinstance(values, (tuple, list)):
            ret = GUESS_SEPARATE[tuple([_.shape[0] for _ in values])]
            if len(ret) == 3:
                space, mask, flavor = ret
                mask_kwargs = {'flavor': flavor, 'legacy': True}
            else:
                space, mask = ret
                mask_kwargs = {}
        else:
            raise TypeError(
                f"`values` has type `{type(values)}, " "which is not supported."
            )
    else:
        boundary = None

    if mask is not None and mask is not False:
        use_mask = True
        if isinstance(mask, (tuple, list)) and all(
            [isinstance(_, np.ndarray) for _ in mask]
        ):
            masks = mask
        else:
            masks = [get_mask(lr, space, **mask_kwargs) for lr in 'lr']
    else:
        use_mask = False

    if isinstance(values, np.ndarray):
        if boundary is not None:
            values = np.array_split(values, boundary)
        elif use_mask:
            values = np.array_split(values, [masks[0].sum()])
        else:
            values = np.split(values, 2)

    if space != "MEBRAIN":
        ico = int(space.split('-ico')[1])
        nv = ico**2 * 10 + 2

    new_values = []
    for v, lr in zip(values, 'lr'):
        if use_mask:
            m = masks['lr'.index(lr)]
            if space == "MEBRAIN":
                nv = {'l': 100988, 'r': 100974}[lr]
            vv = np.full((nv,) + v.shape[1:], np.nan)
            vv[m] = v
        else:
            vv = v

        if space != "MEBRAIN":
            mapping = get_mapping(lr, space, 'MEBRAIN', nn=nn)
            vv = mapping.T @ vv

        new_values.append(vv)

    new_values = np.concatenate(new_values, axis=0)
    return new_values


def to_color(values, cmap, vmax=None, vmin=None):
    if vmin is None:
        vmin = -vmax
    r = (values - vmin) / (vmax - vmin)
    r = np.clip(r, 0.0, 1.0)
    cmap = cm.get_cmap(cmap)
    c = cmap(r)
    return c


def prepare_data(
    values,
    space,
    mask,
    nn=True,
    cmap=None,
    vmax=None,
    vmin=None,
    return_scale=False,
    medial_wall_color=[0.8, 0.8, 0.8, 1.0],
    background_color=[1.0, 1.0, 1.0, 0.0],
):
    values = unmask_and_upsample(values, space, mask, nn=nn)

    if cmap is not None:
        nan_mask = np.isnan(values)
        values = to_color(values, cmap, vmax, vmin)
        values[nan_mask] = medial_wall_color
        values = [
            values,
            [_[: values.shape[1]] for _ in [medial_wall_color, background_color]],
        ]
        values = np.concatenate(values, axis=0)

        if return_scale:
            norm = colors.Normalize(vmax=vmax, vmin=vmin, clip=True)
            scale = cm.ScalarMappable(norm=norm, cmap=cmap)
            return values, scale

    return values


def plot_mebrains(
    values,
    cmap=None,
    vmax=None,
    vmin=None,
    space=None,
    mask=None,
    surf_type='inflated',
    nn=True,
    return_scale=False,
    medial_wall_color=[0.8, 0.8, 0.8, 1.0],
    background_color=[1.0, 1.0, 1.0, 0.0],
    colorbar=True,
    output=None,
    width=500,
    title=None,
    title_size=70,
    fn=None,
    **kwargs,
):
    if output is None and fn is None:
        output = 'pillow'

    assert surf_type in [
        'inflated',
        'pial',
        'midthickness',
        'white',
    ], f"Surface type '{surf_type}' is not recognized."

    if isinstance(values, np.ndarray):
        cat = values
    elif isinstance(values, (tuple, list)):
        cat = np.concatenate(values, axis=0)
    else:
        raise TypeError(
            "Expected `values` to be a numpy array or a list/"
            f"tuple of numpy arrays. Got {type(values)}."
        )
    ndim = cat.ndim
    percentiles = np.nanpercentile(cat, [1, 99])
    max_, min_ = np.nanmax(cat), np.nanmin(cat)

    if ndim not in [1, 2]:
        raise ValueError(f"Expected `values` to be 1D or 2D. Got {ndim}D.")
    if ndim == 1:
        if cmap is None:
            cmap = 'viridis'
        if vmax is None:
            vmax = percentiles[1]
        if vmin is None:
            vmin = percentiles[0]
    elif ndim == 2:
        if max_ > 1 or min_ < 0:
            raise ValueError("Expected `values` to be in [0, 1] when it's 2D.")
        if cat.shape[1] not in [3, 4]:
            raise ValueError(
                "Expected `values` to have 3 or 4 columns (RGB "
                f"or RGBA). Got {cat.shape[1]} columns."
            )

    need_scale = return_scale or colorbar

    ret = prepare_data(
        values,
        space,
        mask,
        nn=nn,
        cmap=cmap,
        vmax=vmax,
        vmin=vmin,
        return_scale=need_scale,
        medial_wall_color=medial_wall_color,
        background_color=background_color,
    )
    if need_scale:
        prepared_values, scale = ret
    else:
        prepared_values = ret

    if surf_type not in PLOT_MAPPING:
        mapping = core_dataset.get(
            os.path.join(
                '2d_plotting_data_mebrains', f'mebrains_to_{surf_type}_image.npy'
            ),
            on_missing='raise',
        )
        PLOT_MAPPING[surf_type] = mapping

    img = prepared_values[PLOT_MAPPING[surf_type]]

    if output == 'raw':
        if return_scale:
            return img, scale
        return img

    img = np.round(img * 255.0).astype(np.uint8)
    if not PIL_ok:
        warn(
            "Skipping conversion to `PIL.Image` because `Pillow` is not "
            "installed. You can install it with `pip install Pillow`."
        )
        if return_scale:
            return img, scale
        return img

    img = PIL_Image.fromarray(img)
    img = Image(img)
    if title is not None:
        img = img.title(title, size=title_size)
    if colorbar:
        img = img.colorbar(scale, **kwargs)

    if fn is not None:
        img.save(fn)

    if return_scale:
        return img, scale
    return img



import numpy as np

import neuroboros as nb


class TestDatasets:
    def test_concat_lr(self):
        dset = nb.Forrest()
        sid = dset.subjects[0]
        for runs in [1, [1, 2, 3]]:
            lh = dset.get_data(sid, 'forrest', runs, 'l')
            rh = dset.get_data(sid, 'forrest', runs, 'r')
            lr = dset.get_data(sid, 'forrest', runs, 'lr')
            cat = np.concatenate([lh, rh], axis=1)
            np.testing.assert_allclose(lr, cat, atol=1e-10)

    def test_concat_lr_scrub(self):
        dset = nb.Forrest(prep='scrub')
        sid = dset.subjects[0]
        for runs in [1, [1, 2, 3]]:
            lh = dset.get_data(sid, 'forrest', 1, 'l')[0]
            rh = dset.get_data(sid, 'forrest', 1, 'r')[0]
            lr = dset.get_data(sid, 'forrest', 1, 'lr')[0]
            cat = np.concatenate([lh, rh], axis=1)
            np.testing.assert_allclose(lr, cat, atol=1e-10)

    def test_concat_runs(self):
        dset = nb.Forrest()
        sid = dset.subjects[0]
        for lr in ['l', 'r', 'lr']:
            cat = np.concatenate(
                [dset.get_data(sid, 'forrest', run_, lr) for run_ in [1, 2, 3]], axis=0
            )
            dm = dset.get_data(sid, 'forrest', [1, 2, 3], lr)
            np.testing.assert_array_equal(dm, cat)

    def test_concat_runs_scrub(self):
        dset = nb.Forrest(prep='scrub')
        sid = dset.subjects[0]
        for lr in ['l', 'r', 'lr']:
            cat = np.concatenate(
                [dset.get_data(sid, 'forrest', run_, lr)[0] for run_ in [1, 2, 3]],
                axis=0,
            )
            mask = np.concatenate(
                [dset.get_data(sid, 'forrest', run_, lr)[1] for run_ in [1, 2, 3]],
                axis=0,
            )
            dm, mask2 = dset.get_data(sid, 'forrest', [1, 2, 3], lr)
            np.testing.assert_array_equal(dm, cat)
            np.testing.assert_array_equal(mask, mask2)



import numpy as np
from scipy.spatial import ConvexHull, cKDTree


def compute_union_coords(coords1, coords2, eps=1e-10):
    nv1 = coords1.shape[0]
    nv2 = coords2.shape[0]
    mask = np.ones((nv2,), dtype=bool)
    indices1 = np.arange(nv1)
    indices2 = np.zeros((nv2,), dtype=int)
    tree = cKDTree(coords1)
    count = nv1
    for i, c in enumerate(coords2):
        d, idx = tree.query(c)
        if d < eps:
            mask[i] = False
            indices2[i] = idx
        else:
            indices2[i] = count
            count += 1

    if not np.any(mask):
        coords = coords1
    else:
        coords = np.concatenate([coords1, coords2[mask]])

    return coords, indices1, indices2


def compute_union_sphere(sphere, coords, eps=1e-10):
    union_coords, indices1, indices2 = compute_union_coords(
        sphere.coords, coords, eps=eps
    )
    if union_coords.shape == sphere.coords.shape:
        return sphere.coords, sphere.faces, indices1, indices2

    nv1 = sphere.coords.shape[0]
    new_coords = union_coords[nv1:]
    t_indices = sphere.barycentric(new_coords, return_sparse=False)[0]

    new_faces = []
    for t_idx in np.unique(t_indices):
        indices = np.where(t_indices == t_idx)[0]
        f = sphere.faces[t_idx]
        cc = np.concatenate([sphere.coords[f], new_coords[indices]], axis=0)
        mp = np.concatenate([f, nv1 + indices])
        hull = ConvexHull(cc)
        for nf in hull.simplices:
            if not np.all(np.isin(nf, np.arange(3))):
                new_faces.append(mp[nf])
    new_faces = np.stack(new_faces)

    mask = np.logical_not(np.isin(np.arange(sphere.faces.shape[0]), t_indices))
    union_faces = np.concatenate([sphere.faces[mask], new_faces], axis=0)

    return union_coords, union_faces, indices1, indices2



import heapq
import warnings
from datetime import datetime

import numpy as np
import scipy.sparse as sparse
from scipy.spatial.distance import cdist


def subdivide_edges(coords, faces, n_div):
    """Subdivide each edge into ``n_div`` parts.

    This function adds new vertices to the mesh. The new vertices are placed
    on each edge of the mesh. The number of new vertices on each edge is
    ``n_div - 1``, which means each edge is now divided into ``n_div`` parts.

    The new vertices are usually used to increase the accuracy of the
    estimated geodesic distances using Dijkstra's algorithm.

    Parameters
    ----------
    coords : ndarray of shape (nv, 3)
        The coordinates of the vertices.
    faces : ndarray of shape (nf, 3)
        The indices of the vertices of each triangle face.
    n_div : int
        The number of segments to divide each edge into.


    Returns
    -------
    new_coords : ndarray of shape (nv_new, 3)
        The coordinates of the new vertices. It should be concatenated to the
        original coordinates to get the full set of vertex coordinates.
    e_mapping : dict
        The mapping from the original edges to the new vertices. The keys are
        tuples of sorted vertex indices of the original edges. The values are
        the indices of the new vertices. Vertex indices start from ``nv``.
    neighbors : dict
        The neighbors of each vertex. The keys are the vertex indices. The
        values are dictionaries. The keys of the dictionaries are the indices
        of the neighboring vertices. The values of the dictionaries are the
        distances between the vertex and the neighboring vertices.
    """
    n_edges = faces.shape[0] * 3 // 2
    nv_new = n_edges * (n_div - 1)
    new_coords = np.zeros((nv_new, 3), dtype=coords.dtype)
    print(new_coords.shape)
    neighbors = {}

    nv = coords.shape[0]
    count = 0
    e_mapping = {}
    arng = np.arange(1, n_div)
    for f in faces:
        for i in f:
            if i not in neighbors:
                neighbors[i] = {}

        steiner = []
        for i, j, k in [[0, 1, 2], [1, 2, 0], [2, 0, 1]]:
            a, b, c = f[[i, j, k]]
            e = (a, b) if a < b else (b, a)
            if e not in e_mapping:
                cc = (
                    coords[[e[1]]] * arng[:, np.newaxis]
                    + coords[e[0]] * (n_div - arng[:, np.newaxis])
                ) / n_div
                new_coords[count : count + n_div - 1] = cc
                indices = (count + nv - 1) + arng
                e_mapping[e] = indices
                count += n_div - 1
            else:
                indices = e_mapping[e]
                cc = new_coords[indices - nv]
            steiner.append([cc, indices])

            # connecting points on an edge and the opposite point
            idx2, c2 = c, coords[c]
            for idx1, c1 in zip(indices, cc):
                if idx1 not in neighbors:
                    neighbors[idx1] = {}
                if idx1 not in neighbors[idx2]:
                    d = np.linalg.norm(c1 - c2)
                    neighbors[idx2][idx1] = d
                    neighbors[idx1][idx2] = d

        # connecting points on an edge and points on another edge
        for i, j in [[0, 1], [1, 2], [2, 0]]:
            cc1, indices1 = steiner[i]
            cc2, indices2 = steiner[j]
            for idx1, c1 in zip(indices1, cc1):
                for idx2, c2 in zip(indices2, cc2):
                    if idx1 not in neighbors[idx2]:
                        d = np.linalg.norm(c1 - c2)
                        neighbors[idx2][idx1] = d
                        neighbors[idx1][idx2] = d

        # connecting points of the original triangle
        for i in f:
            for j in f:
                if i != j and i not in neighbors[j]:
                    d = np.linalg.norm(coords[i] - coords[j])
                    neighbors[i][j] = d
                    neighbors[j][i] = d

    return new_coords, e_mapping, neighbors


def dijkstra_distances(nv, candidates, neighbors, max_dist=None):
    """Dijkstra's algorithm to estimate the geodesic distances.

    This function estimates the geodesic distances between a vertex and all
    vertices in the mesh using Dijkstra's algorithm. If the vertex is part of
    the mesh, the ``candidates`` list should contain only one tuple
    ``(0.0, idx)`` where ``idx`` is the index of the vertex. If the vertex is
    not part of the mesh, the ``candidates`` list should contain tuples
    ``(d, idx)`` where ``d`` is the distance between the vertex and the
    neighboring vertices and ``idx`` is the index of the neighboring vertices.

    Parameters
    ----------
    nv : int
        The number of vertices.
    candidates : list
        The list of candidate vertices. Each element is a tuple of the
        distance and the vertex index.
    neighbors : dict
        The neighbors of each vertex. The keys are the vertex indices. The
        values are dictionaries. The keys of the dictionaries are the indices
        of the neighboring vertices. The values of the dictionaries are the
        distances between the vertex and the neighboring vertices.
    max_dist : float or None, default=None
        The maximum distance to search. If None, the maximum distance is
        infinity.

    Returns
    -------
    dists : ndarray of shape (nv,)
        The estimated geodesic distances.
    """
    dists = np.full((nv,), np.inf)
    finished = np.zeros((nv,), dtype=bool)
    for d, idx in candidates:
        dists[idx] = d

    while candidates:
        d, idx = heapq.heappop(candidates)
        if finished[idx]:
            continue

        for nbr, nbr_d in neighbors[idx].items():
            new_d = d + nbr_d
            if max_dist is not None and new_d > max_dist:
                continue
            if new_d < dists[nbr]:
                dists[nbr] = new_d
                heapq.heappush(candidates, (new_d, nbr))
        finished[idx] = True

    return dists


def subdivision_voronoi(
    coords,
    faces,
    e_mapping,
    neighbors,
    f_indices,
    weights,
    max_dist=None,
    verbose=False,
):
    """Voronoi diagram on a subdivided mesh.

    This function estimates the nearest vertex on the target mesh for each
    vertex of the subdivided mesh. The subdivided mesh is used to increase the
    accuracy of the estimated geodesic distances using Dijkstra's algorithm.

    Barycentric interpolation is used to map the vertices of the target mesh
    to the subdivided mesh.

    Parameters
    ----------
    coords : ndarray of shape (nv, 3)
        The coordinates of the vertices of the subdivided mesh.
    faces : ndarray of shape (nf, 3)
        The vertex indices of each triangle face of the original mesh.
    e_mapping : dict
        The mapping from the original edges to the new vertices. The keys are
        tuples of sorted vertex indices of the original edges. The values are
        the indices of the new vertices.
    neighbors : dict
        The neighbors of each vertex. The keys are the vertex indices. The
        values are dictionaries. The keys of the dictionaries are the indices
        of the neighboring vertices. The values of the dictionaries are the
        distances between the vertex and the neighboring vertices.
    f_indices : ndarray of shape (nv_target,)
        Each element is the index of the triangle face on the original mesh
        that contains the corresponding vertex on the target mesh.
    weights : ndarray of shape (nv_target, 3)
        Each row is the barycentric weights of the three vertices of the face
        for the corresponding vertex on the target mesh.
    max_dist : float or None, default=None
        The maximum distance of the initial search. If None, the maximum
        distance is computed based on the number of vertices on the target
        mesh.
    verbose : bool, default=False
        Whether to print the progress or not.

    Returns
    -------
    nn : ndarray of shape (nv,)
        The indices of the nearest vertex on the target mesh for each vertex
        of the subdivided mesh.
    nnd : ndarray of shape (nv,)
        The estimated geodesic distances between each vertex of the subdivided
        mesh and the nearest vertex on the target mesh.
    """
    nv = coords.shape[0]
    n1, n2 = len(np.unique(f_indices)), len(f_indices)
    if n1 != n2:
        warnings.warn(f"{n2} template vertices are mapped to {n1} faces.")
    if max_dist is None:
        max_dist = 4.0 * np.sqrt(10242 / f_indices.shape[0]) + 2.0
    log_step = f_indices.shape[0] // 100 + 1
    nn = np.full((nv,), -1, dtype=int)
    nnd = np.full((nv,), np.inf)
    while np.any(np.isinf(nnd)):
        for i, (f_idx, w) in enumerate(zip(f_indices, weights)):
            cc = w @ coords[faces[f_idx]]
            a, b, c = sorted(faces[f_idx])
            indices = np.concatenate(
                [e_mapping[(a, b)], e_mapping[(a, c)], e_mapping[(b, c)], [a, b, c]]
            )
            dd = cdist(cc[np.newaxis], coords[indices], 'euclidean').ravel()
            candidates = []
            for d, idx in zip(dd, indices):
                heapq.heappush(candidates, (d, idx))
            d = dijkstra_distances(nv, candidates, neighbors, max_dist=max_dist)
            mask = d < nnd
            nn[mask] = i
            nnd[mask] = d[mask]
            if verbose and i % log_step == 0:
                print(
                    datetime.now(),
                    i,
                    mask.sum(),
                    np.isfinite(d).sum(),
                    d.shape,
                    d.max(),
                    d.min(),
                    len(candidates),
                    np.isinf(nnd).sum(),
                )
        max_dist *= 1.5
    print(nnd.max())
    return nn, nnd


def native_voronoi(coords, faces, e_mapping, neighbors, verbose=False):
    """Voronoi diagram on the original mesh.

    This function estimates the nearest vertex on the original mesh for each
    vertex of the subdivided mesh. The subdivided mesh is used to increase the
    accuracy of the estimated geodesic distances using Dijkstra's algorithm.

    Parameters
    ----------
    coords : ndarray of shape (nv, 3)
        The coordinates of the vertices of the subdivided mesh.
    faces : ndarray of shape (nf, 3)
        The vertex indices of each triangle face of the original mesh.
    e_mapping : dict
        The mapping from the original edges to the new vertices. The keys are
        tuples of sorted vertex indices of the original edges. The values are
        the indices of the new vertices.
    neighbors : dict
        The neighbors of each vertex. The keys are the vertex indices. The
        values are dictionaries. The keys of the dictionaries are the indices
        of the neighboring vertices. The values of the dictionaries are the
        distances between the vertex and the neighboring vertices.
    verbose : bool, default=False
        Whether to print the progress or not.

    Returns
    -------
    nn : ndarray of shape (nv,)
        The indices of the nearest vertex on the original mesh for each vertex
        of the subdivided mesh.
    nnd : ndarray of shape (nv,)
        The estimated geodesic distances between each vertex of the subdivided
        mesh and the nearest vertex on the original mesh.
    """
    nv = coords.shape[0]
    nn = np.full((nv,), -1, dtype=int)
    nnd = np.full((nv,), np.inf)
    max_dist = np.max(
        [
            np.linalg.norm(coords[faces[:, 0]] - coords[faces[:, 1]], axis=1).max(),
            np.linalg.norm(coords[faces[:, 1]] - coords[faces[:, 2]], axis=1).max(),
            np.linalg.norm(coords[faces[:, 2]] - coords[faces[:, 0]], axis=1).max(),
        ]
    )
    max_dist = max_dist * 0.5 + 1e-3
    print(max_dist)

    seeds = []

    for f in faces:
        f = np.sort(f)
        cc1 = coords[f]
        a, b, c = f
        for i, e in enumerate([(b, c), (a, c), (a, b)]):
            indices = e_mapping[e]
            cc2 = coords[indices]
            d = cdist(cc1, cc2)

            min_idx = np.argmin(d, axis=0)
            # wide short triangle
            if np.any(min_idx == i):
                seeds.append(f[i])
            d = d.min(axis=0)
            mask = d < nnd[indices]

            nnd[indices[mask]] = d[mask]
            nn[indices[mask]] = f[min_idx[mask]]
        nn[f] = f
        nnd[f] = 0.0
    if verbose:
        print(np.isfinite(nnd).mean(), nnd.max())

    seeds = np.unique(seeds)
    if verbose:
        print(len(seeds), seeds[:10])
    for i, seed in enumerate(seeds):
        candidates = [(0.0, seed)]
        d = dijkstra_distances(nv, candidates, neighbors, max_dist=max_dist)
        mask = d < nnd
        nn[mask] = seed
        nnd[mask] = d[mask]
        if verbose and i % 10000 == 0:
            print(
                datetime.now(),
                i,
                seed,
                mask.sum(),
                nnd.max(),
                np.isfinite(d).sum(),
                d.shape,
                d.max(),
                d.min(),
                np.isinf(nnd).sum(),
            )

    return nn, nnd


def inverse_face_mapping(f_indices, weights, coords, faces):
    """Inverse face mapping.

    This function computes the mapping from the original faces to the target
    vertices. Given a triangle face on the original mesh, the mapping returns
    the indices of the target vertices inside the triangle face and the
    coordinates of the target vertices.

    Parameters
    ----------
    f_indices : ndarray of shape (nv_target,)
        Each element is the index of the triangle face on the original mesh
        that contains the corresponding vertex on the target mesh.
    weights : ndarray of shape (nv_target, 3)
        Each row is the barycentric weights of the three vertices of the face
        for the corresponding vertex on the target mesh.
    coords : ndarray of shape (nv, 3)
        The coordinates of the vertices of the original mesh.
    faces : ndarray of shape (nf, 3)
        The vertex indices of each triangle face of the original mesh.

    Returns
    -------
    f_inv : dict
        The inverse mapping from the original faces to the target vertices.
        The keys are the indices of the original faces. The values are lists
        of pairs of the indices of the target vertices and the coordinates of
        the target vertices.
    """
    f_inv = {}
    for i, f_idx in enumerate(f_indices):
        if f_idx not in f_inv:
            f_inv[f_idx] = []
        f_inv[f_idx].append([i, weights[i] @ coords[faces[f_idx]]])
    return f_inv


def split_triangle(t_div):
    ww1 = []
    for i in range(t_div):
        for j in range(t_div - i):
            k = t_div - i - j - 1
            ww1.append([i, j, k])
    ww1 = (np.array(ww1) + 1.0 / 3) / t_div
    ww2 = []
    for i in range(t_div - 1):
        for j in range(t_div - i - 1):
            k = t_div - i - j - 1
            ww2.append([i, j, k])
    ww2 = (np.array(ww2) + np.array([[2 / 3, 2 / 3, -1 / 3]])) / t_div
    ww = np.concatenate([ww1, ww2])
    return ww


def compute_occupation(f_idx, f, coords, indices, nn, nnd, f_inv, ww):
    """
    Parameters
    ----------
    f_idx : int
        The index of the triangular face on the original mesh.
    f : ndarray of shape (3,)
        The vertex indices of the triangular face.
    coords : ndarray of shape (nv, 3)
        The coordinates of the vertices of the subdivided mesh.
    indices : ndarray of shape (n,)
        The indices of all vertices of the subdivided mesh that are on the
        edges of the triangular face.
    nn : ndarray of shape (nv,)
        The indices of the nearest vertex on the target mesh for each vertex
        of the subdivided mesh.
    nnd : ndarray of shape (nv,)
        The geodesic distances between each vertex of the subdivided mesh and
        the nearest vertex on the target mesh.
    f_inv : dict
        The inverse mapping from the original faces to the target vertices.
    ww : ndarray of shape (m, 3)
        The barycentric weights of the points on the triangular face. Usually
        the output of the ``split_triangle`` function.

    Returns
    -------

    """
    nn1 = [nn[indices]]
    u = np.unique(nn1)

    # The same vertex of the target mesh is the nearest for all vertices on
    # the edges of the triangular face.
    if len(u) == 1 and len(f_inv) == 0:
        return {u[0]: np.ones((ww.shape[0],), dtype=bool)}

    cc1 = [coords[indices]]
    dd1 = [nnd[indices]]
    # Vertices of the target mesh that are within the triangular face.
    if f_idx in f_inv:
        for i, c in f_inv[f_idx]:
            cc1.append([c])
            dd1.append([0.0])
            nn1.append([i])
    cc1 = np.concatenate(cc1)
    dd1 = np.concatenate(dd1)
    nn1 = np.concatenate(nn1)
    cc2 = ww @ coords[f]
    d = cdist(cc1, cc2, 'euclidean') + dd1[:, np.newaxis]
    idx = np.argmin(d, axis=0)
    occupation = nn1[idx]
    return {u: occupation == u for u in np.unique(occupation)}


def compute_overlap(
    faces,
    face_areas,
    e_mapping,
    coords,
    nn,
    nnd,
    f_inv,
    nn2,
    nnd2,
    f_inv2,
    nv1,
    nv2,
    t_div=32,
):
    mat = sparse.lil_matrix((nv1, nv2))
    ww = split_triangle(t_div)
    for f_idx, f in enumerate(faces):
        a, b, c = sorted(faces[f_idx])
        indices = np.concatenate(
            [e_mapping[(a, b)], e_mapping[(a, c)], e_mapping[(b, c)], [a, b, c]]
        )
        uu1 = compute_occupation(f_idx, f, coords, indices, nn, nnd, f_inv, ww)
        uu2 = compute_occupation(f_idx, f, coords, indices, nn2, nnd2, f_inv2, ww)
        for u1, m1 in uu1.items():
            for u2, m2 in uu2.items():
                overlap = np.logical_and(m1, m2).mean()
                if overlap:
                    mat[u1, u2] += overlap * face_areas[f_idx]
    return mat.tocsr()


def overlap_transform(
    anat,
    sphere,
    tpl_coords,
    coords,
    e_mapping,
    neighbors,
    native_nn,
    native_nnd,
    t_div=32,
):
    f_indices, weights = sphere.barycentric(tpl_coords, eps=1e-7, return_sparse=False)
    nn, nnd = subdivision_voronoi(
        coords, anat.faces, e_mapping, neighbors, f_indices, weights
    )
    f_inv = inverse_face_mapping(f_indices, weights, anat.coords, anat.faces)
    mat = compute_overlap(
        anat.faces,
        anat.face_areas,
        e_mapping,
        coords,
        native_nn,
        native_nnd,
        {},
        nn,
        nnd,
        f_inv,
        anat.nv,
        tpl_coords.shape[0],
        t_div=t_div,
    )
    return mat


"""
Similar to https://github.com/mojocorp/geodesic subdivision
"""



"""
==============================
Neuroboros (:mod:`neuroboros`)
==============================

"""

from . import linalg as linalg
from . import stats as stats
from . import surface as surface
from ._version import __version__, __version_tuple__

# from . import idm as idm
from .benchmark import classification
from .datasets import (
    HBNSSI,
    IBC,
    ID1000,
    Bologna,
    Budapest,
    CamCAN,
    Dalmatians,
    Forrest,
    Life,
    MonkeyKingdom,
    Raiders,
    SpaceTop,
    Whiplash,
)
from .datasets import get_dataset as dataset
from .glm import glm
from .plot2d import Image
from .plot2d import brain_plot as plot
from .plot_mebrains import plot_mebrains as plot_mebrains
from .searchlights import get_searchlights as sls
from .spaces import get_distances as distances
from .spaces import get_geometry as geometry
from .spaces import get_mapping as mapping
from .spaces import get_mask as mask
from .spaces import get_morphometry as morphometry
from .spaces import get_parcellation as parcellation
from .spaces import smooth as smooth
from .surface import Sphere, Surface
from .utils import load, percentile, save
from .utils import save_results
from .utils import save_results as record



"""
===================================
Utilities (:mod:`neuroboros.utils`)
===================================

.. currentmodule:: neuroboros.utils

Function execution
==================

.. autosummary::
    :toctree:

    monitor - Monitors function execution and records running time and system information.
    save_results - Run the function and save the returned output to file.
    parse_record - Parse the timing information based on record.

Input / Output
==============

.. autosummary::
    :toctree:

    save - Save the data using the automatically determined format.
    load - Load the file using the automatically determined function.

High-performance computing
==========================

.. autosummary::
    :toctree:

    assert_sufficient_time - Check if remaining SLURM walltime is sufficient.

"""

import functools
import gzip
import json
import os
import pickle
import subprocess
import time
import warnings
from datetime import datetime, timedelta

import joblib
import nibabel as nib
import numpy as np
import pandas as pd
import scipy.sparse as sparse

try:
    from PIL import Image

    PIL_ok = True
except ImportError as e:
    PIL_ok = False


def percentile(data, ignore_nan=False, **kwargs):
    count = np.isnan(data).sum()
    if count:
        if not ignore_nan:
            if not isinstance(data, np.ndarray):
                data = np.asarray(data)
            warnings.warn(f'{count} values out of {data.size} are NaNs.')
        print(np.nanpercentile(data, np.linspace(0, 100, 11), **kwargs))
    else:
        print(np.percentile(data, np.linspace(0, 100, 11), **kwargs))


def save(fn, data):
    """Save the data using the automatically determined format.

    If ``fn`` ends with ".npy", save as npy file.
    If ``data`` is spmatrix, save as npz file.
    If dict and ``fn`` ends with ".npz", save as npz file.
    If ``fn`` ends with ".pkl", save using ``pickle.dump``.
    If ``fn`` ends with ".png", save as PNG file (requires the Pillow
    package).
    If ``fn`` ends with ".json", save as JSON file.

    The function also automatically create the directory ``fn`` is in if it
    does not exist.

    Parameters
    ----------
    fn : str
        The filename (including path) to save the data to.
    data : ndarray or spmatrix or dict or object
        The data to be saved.

    Returns
    -------
    ret
        Returned value of ``numpy.save``, ``numpy.savez``,
        ``scipy.sparse.save_npz``, or ``pickle.dump``.

    Raises
    ------
    TypeError
        Raised when ``type(data)`` is not supported.


    """
    dirname = os.path.dirname(fn)
    if dirname:
        os.makedirs(dirname, exist_ok=True)

    if fn.endswith('.npy'):
        if not isinstance(data, np.ndarray):
            warnings.warn('`data` is not an ndarray, trying to convert.')
        return np.save(fn, data)

    if sparse.isspmatrix(data):
        return sparse.save_npz(fn, data)
    if fn.endswith('.npz') and isinstance(data, dict):
        return np.savez(fn, **data)

    if fn.endswith('.pkl'):
        with open(fn, 'wb') as f:
            return pickle.dump(data, f)

    if fn.endswith('.png'):
        assert PIL_ok, "Needs the Pillow package to save images."
        if isinstance(data, Image.Image):
            im = data
        elif isinstance(data, np.ndarray):
            im = Image.fromarray(data)
        else:
            raise TypeError(f"Cannot save '{type(data)}' to image.")
        return im.save(fn)

    if fn.endswith('.shape.gii'):
        darray = nib.gifti.GiftiDataArray(
            data.astype(np.float32),
            intent=nib.nifti1.intent_codes['NIFTI_INTENT_SHAPE'],
            datatype=nib.nifti1.data_type_codes['NIFTI_TYPE_FLOAT32'],
        )
        gii = nib.gifti.GiftiImage(darrays=[darray])
        return nib.save(gii, fn)

    if fn.endswith('.json'):
        with open(fn, 'w') as f:
            return json.dump(data, f)

    raise TypeError(f'`data` type {type(data)} not supported.')


def load(fn, **kwargs):
    """Load the file using the automatically determined function.

    Parameters
    ----------
    fn : str
        File name of the file to be loaded.

    Returns
    -------
    ret
        Returned value of ``numpy.load``, ``scipy.sparse.load_npz``,
        ``pickle.load``, ``pandas.read_csv``, or ``json.load``.

    Raises
    ------
    TypeError
        Raised when the type of ``fn`` is not supported.
    """
    if fn.endswith('.npy'):
        return np.load(fn, **kwargs)
    if fn.endswith('.npz'):
        try:
            return sparse.load_npz(fn, **kwargs)
        except (OSError, ValueError):
            return np.load(fn, **kwargs)
    if fn.endswith('.tsv'):
        d = dict(delimiter='\t', na_values='n/a')
        d.update(kwargs)
        return pd.read_csv(fn, **d)
    if fn.endswith('.csv'):
        d = dict(na_values='n/a')
        d.update(kwargs)
        return pd.read_csv(fn, **d)
    if fn.endswith('.pkl'):
        with open(fn, 'rb') as f:
            return pickle.load(f, **kwargs)
    if fn.endswith('.json.gz'):
        with gzip.open(fn, 'rb') as f:
            return json.load(f, **kwargs)
    if fn.endswith('.json'):
        with open(fn, 'rb') as f:
            return json.load(f, **kwargs)
    raise TypeError(f'file type of `fn` is not supported.')


def parse_record(record_fn, assert_node=None):
    """Parse the timing information based on record.

    This function parses the record file generated by ``monitor`` or
    ``save_results`` and returns the elapsed CPU time and wall time as a NumPy
    array.

    Parameters
    ----------
    record_fn : str
        File name of the record file.

    Returns
    -------
    t : ndarray
        A numpy array with 2 elements, where the 1st is the CPU time, and the
        2nd is the wall time. Both are in nanoseconds.
    """
    with open(record_fn) as f:
        lines = f.read().splitlines()
    cpu_time = int(lines[2]) - int(lines[1])
    wall_time = int(lines[4]) - int(lines[3])
    if assert_node is not None:
        if assert_node not in [lines[5], lines[5].split('.')[0]]:
            raise ValueError(
                f"Expecting Node `{assert_node}`, got {lines[5]} in record."
            )
    t = np.array([cpu_time, wall_time])
    return t


def monitor(func, record_fn=None):
    """
    Monitors function execution and records running time and system information.

    This function takes a function ``func`` as input and outputs a wrapped
    function ``wrapped_func``. ``wrapped_func`` is similar to ``func``, except
    that it also records the running time of function execution and system
    information. When ``record_fn`` is None, ``wrapped_func`` will return a
    tuple, where the 1st element is the record and the 2nd element is the
    original returned value of ``func``. When it's not None, the record will
    be saved into the file, and only the original returned value of ``func``
    will be returned.

    Parameters
    ----------
    func : function
        The original function to be wrapped.
    record_fn : {str, None}, default=None
        When it is None, the monitoring record will be the first returned
        value of ``wrapped_func``. Otherwise, the information will be saved
        into a text file with the file name ``record_fn`` and only the
        original returned value of ``func`` will be returned.

    Returns
    -------
    wrapped_func : function
        The wrapped function that will record running information.
    """
    fmt = '%Y-%m-%d %H:%M:%S.%f'

    def monitored_func(*args, **kwargs):
        cpu_time_start = time.process_time_ns()
        wall_time_start = time.perf_counter_ns()
        results = func(*args, **kwargs)
        cpu_time_end = time.process_time_ns()
        wall_time_end = time.perf_counter_ns()

        hostname = os.uname()[1]
        total_cpus = os.cpu_count()
        avail_cpus = joblib.cpu_count()

        info = (
            f'Computation finished at {datetime.now().strftime(fmt)}\n'
            f'{cpu_time_start}\n{cpu_time_end}\n'
            f'{wall_time_start}\n{wall_time_end}\n'
            f'{hostname}\n{avail_cpus}\n{total_cpus}\n'
        )
        if record_fn is None:
            return info, results
        else:
            with open(record_fn, 'w') as f:
                f.write(info)
            return results

    wrapped_func = functools.wraps(func)(monitored_func)
    return wrapped_func


def save_results(
    out_fn,
    func,
    return_results=False,
    log_fn=None,
    rerun_hours=48,
    verbose=True,
    rerun=False,
):
    """Run the function and save the returned output to file.

    This function takes a function ``func`` as input and outputs a wrapped
    function ``wrapped_func``. ``wrapped_func`` is similar to ``func``, except
    that it will also save the returned output of ``func`` to files (file
    names indicated by ``out_fn``), so that it can be accessed later.

    Parameters
    ----------
    out_fn : str or list/tuple of str
        The names of the output file(s).
    func : function
        The function to be wrapped.
    return_results : bool, default=False
        Whether ``wrapped_func`` will return the output of the original
        function ``func``. If it's False, None will be returned.
    log_fn : {str, None}, default=None
        The root file name of the text files for logging.
    rerun_hours : int, default=48
        Rerun the function if the start of the last run was more than
        ``rerun_hours`` ago.
    verbose : bool, default=True
        Whether to output additional logging information.
    rerun : bool, default=False
        Whether to rerun the function even if the output files exist.

    Returns
    -------
    wrapped_func : function
        The wrapped function that will run ``func`` and save the results to
        ``out_fn``.

    Raises
    ------
    ValueError
        Raised when ``out_fn`` is neither a string or a tuple/list of strings.
    """
    if isinstance(out_fn, (list, tuple)):
        out_fns = out_fn
    elif isinstance(out_fn, str):
        out_fns = [out_fn]
    else:
        raise ValueError("`out_fn` must be a string or a list/tuple of strings.")

    if log_fn is None:
        log_fn = out_fns[0]

    running_fn = log_fn + '.running'
    finish_fn = log_fn + '.finish'
    fmt = '%Y-%m-%d %H:%M:%S.%f'

    monitored_func = monitor(func, finish_fn)

    def func_w_cache(*args, **kwargs):
        if not rerun:
            all_exist = False

            if os.path.exists(running_fn):
                while True:
                    with open(running_fn) as f:
                        diff = datetime.now() - datetime.strptime(f.read(), fmt)
                    if diff < timedelta(hours=rerun_hours):
                        if not return_results:
                            return
                        time.sleep(600)
                    else:
                        break

            if os.path.exists(finish_fn):
                all_exist = True
                if verbose:
                    print(datetime.now(), f'`finish_fn` exists: {finish_fn}')
            elif all([os.path.exists(_) for _ in out_fns]):
                all_exist = True
                if verbose:
                    print(datetime.now(), f'All output files exist: {out_fns}')

            if all_exist:
                if not return_results:
                    return
                else:
                    results = [load(_) for _ in out_fns]
                    if len(results) == 1:
                        return results[0]
                    else:
                        return results

        os.makedirs(os.path.dirname(log_fn), exist_ok=True)
        with open(running_fn, 'w') as f:
            f.write(datetime.now().strftime(fmt))

        if verbose:
            print(datetime.now(), f'Starting to compute for: {out_fns}')

        results = monitored_func(*args, **kwargs)

        if len(out_fns) > 1:
            for res, fn in zip(results, out_fns):
                save(fn, res)
        else:
            save(out_fns[0], results)

        if os.path.exists(running_fn):
            os.remove(running_fn)

        if return_results:
            return results
        return

    wrapped_func = functools.wraps(func)(func_w_cache)
    return wrapped_func


def assert_sufficient_time(minimum='1:00:00'):
    """Check if remaining SLURM walltime is sufficient.

    This function checks if the remaining walltime of SLURM is larger than the
    ``minimum`` when the script is run through the SLURM job manager. If the
    remaining walltime is insufficient, it will stop the remaining parts of
    the script from running by executing ``exit(0)``.

    Parameters
    ----------
    minimum : str
        The minimal walltime in SLURM format. E.g., '1-02:03:04' means 1 day,
        2 hours, 3 minutes, and 4 seconds.
    """
    if 'SLURM_JOBID' not in os.environ:
        return

    cmd = ['squeue', '-h', '-j', os.environ['SLURM_JOBID'], '-o', '"%L"']
    sp = subprocess.run(cmd, capture_output=True)
    remaining = sp.stdout.decode('ascii').split('"')[1]

    def parse_time(s):
        days = int(s.split('-')[0]) if '-' in s else 0
        parts = s.split('-')[-1].split(':')
        hours = int(parts[-3]) if len(parts) >= 3 else 0
        minutes = int(parts[-2]) if len(parts) >= 2 else 0
        seconds = int(parts[-1]) if len(parts) >= 1 else 0
        return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)

    r = parse_time(remaining)
    m = parse_time(minimum)
    print(datetime.now(), 'Remaining walltime:', r)
    print(datetime.now(), 'Minimal walltime:  ', m)
    if r < m:
        print(datetime.now(), 'Insufficient remaining walltime time. Exiting.')
        exit(0)


def optimize_dtype(data):
    """Optimize the data type of a NumPy array for memory efficiency.

    Parameters
    ----------
    data : ndarray
        The input NumPy array to be optimized.

    Returns
    -------
    optimized_data : ndarray
        The NumPy array with the optimized data type.

    Raises
    ------
    ValueError
        If the input data is not a NumPy array.
    TypeError
        If the input data is a scalar value.
    """
    if not isinstance(data, np.ndarray):
        raise ValueError("Input data must be a numpy array.")
    if data.ndim == 0:
        raise TypeError("Input data must be a non-scalar numpy array.")
    optimized_data = data
    for dtype in [np.int32, np.uint32, np.int16, np.uint16, np.int8, np.uint8]:
        if np.all(data.astype(dtype) == data):
            optimized_data = data.astype(dtype)
    return optimized_data



"""
================================================
Neuroboros datasets (:mod:`neuroboros.datasets`)
================================================

.. currentmodule:: neuroboros.datasets

.. autosummary::
    :toctree:

    Forrest - The StudyForrest 3 T dataset.


"""

import os
from functools import partial
from glob import glob

import numpy as np
from scipy.stats import zscore

from ..io import DatasetManager
from ..spaces import get_mask

SURFACE_SPACES = ['fsavg-ico32', 'onavg-ico32', 'onavg-ico48', 'onavg-ico64']
SURFACE_RESAMPLES = [
    '1step_pial_overlap',
    '1step_pial_area',
    '2step_normals-equal_nnfr',
    '2step_normals-sine_nnfr',
]
VOLUME_SPACES = ['mni-2mm', 'mni-3mm', 'mni-4mm']
VOLUME_RESAMPLES = ['1step_linear_overlap', '1step_fmriprep_overlap']


def guess_surface_volume(space, resample, lr):
    if space in SURFACE_SPACES or resample in SURFACE_RESAMPLES:
        return 'surface'
    if space in VOLUME_SPACES or resample in VOLUME_RESAMPLES:
        return 'volume'
    if lr in ['l', 'r', 'l-cerebrum', 'r-cerebrum', 'lr']:
        return 'surface'
    return 'volume'


def default_prep(dm, confounds, cortical_mask, z=True, mask=True, gsr=False):
    if mask and cortical_mask is not None:
        dm = dm[:, cortical_mask]
    conf = confounds[0]
    if gsr:
        gs = np.array(confounds[1]['global_signal'])
        conf = np.concatenate([conf, gs[:, np.newaxis]], axis=1)
    finite_mask = np.all(np.isfinite(dm), axis=0)
    beta = np.linalg.lstsq(conf, dm[:, finite_mask], rcond=None)[0]
    dm[:, finite_mask] = dm[:, finite_mask] - conf @ beta
    if z:
        dm = np.nan_to_num(zscore(dm, axis=0))
    return dm


def scrub_prep(dm, confounds, cortical_mask, z=True, mask=True, gsr=False):
    if mask and cortical_mask is not None:
        dm = dm[:, cortical_mask]
    conf, _, keep = confounds
    if gsr:
        gs = np.array(confounds[1]['global_signal'])
        conf = np.concatenate([conf, gs[:, np.newaxis]], axis=1)
    dm = dm[keep]
    finite_mask = np.all(np.isfinite(dm), axis=0)
    beta = np.linalg.lstsq(conf[keep], dm[:, finite_mask], rcond=None)[0]
    dm[:, finite_mask] = dm[:, finite_mask] - conf[keep] @ beta
    if z:
        dm = np.nan_to_num(zscore(dm, axis=0))
    return dm, keep


def get_prep(name, **kwargs):
    if name.endswith('-gsr'):
        gsr = True
        name = name[:-4]
    else:
        gsr = False

    prep = {
        'default': default_prep,
        'scrub': scrub_prep,
    }[name]
    if gsr:
        prep = partial(prep, gsr=True)
    if kwargs:
        prep = partial(prep, **kwargs)
    return prep


class Dataset:
    def __init__(
        self,
        name,
        dl_source,
        root_dir,
        space,
        resample,
        surface_space=None,
        surface_resample=None,
        volume_space=None,
        volume_resample=None,
        prep='default',
        fp_version='20.2.7',
    ):
        self.name = name

        self.dl_source = dl_source
        self.root_dir = root_dir

        self.dl_dset = DatasetManager(
            self.name, root=self.root_dir, source=self.dl_source
        )

        # if self.dl_source is None:
        #     if self.root_dir is not None:
        #         self.dl_dset = LocalDataset(self.name, self.root_dir)
        #     else:
        #         try:
        #             self.dl_dset = LocalDataset(self.name, self.root_dir)
        #         except AssertionError as e:
        #             raise RuntimeError(
        #                 "Dataset not found locally and `dl_source` not " "specified."
        #             ) from e
        # else:
        #     self.dl_dset = DefaultDataset(self.name, self.dl_source, self.root_dir)

        self.fp_version = fp_version
        self.surface_space = surface_space
        self.volume_space = volume_space
        self.surface_resample = surface_resample
        self.volume_resample = volume_resample

        if space is not None:
            if not isinstance(space, (tuple, list)):
                space = [space]
            for sp in space:
                if sp in SURFACE_SPACES and self.surface_space is None:
                    self.surface_space = sp
                elif sp in VOLUME_SPACES and self.volume_space is None:
                    self.volume_space = sp
                else:
                    raise ValueError(f"space {sp} not recognized.")

        if resample is not None:
            if not isinstance(resample, (tuple, list)):
                resample = [resample]
            for resamp in resample:
                if resamp in SURFACE_RESAMPLES and self.surface_resample is None:
                    self.surface_resample = resamp
                elif resamp in VOLUME_RESAMPLES and self.volume_resample is None:
                    self.volume_resample = resamp
                else:
                    raise ValueError(f"Resampling method {resamp} not recognized.")

        self.prep = prep

        self.renaming = self.dl_dset.get('rename.json.gz')

        self.subject_sets = {}
        for fn in sorted(
            glob(os.path.join(self.dl_dset.root, 'subject_sets', '*.txt'))
        ):
            task = os.path.basename(fn)[:-4]
            with open(fn) as f:
                self.subject_sets[task] = f.read().splitlines()

    def load_data(self, sid, task, run, lr, space, resample, fp_version=None):
        if lr == 'lr':
            dm = np.concatenate(
                [
                    self.load_data(sid, task, run, lr_, space, resample, fp_version)
                    for lr_ in 'lr'
                ],
                axis=1,
            )
            return dm

        if fp_version is None:
            fp_version = self.fp_version

        if lr in ['l', 'r']:
            lr = f'{lr}-cerebrum'

        if self.renaming is None:
            fn = [
                fp_version,
                'resampled',
                space,
                lr,
                resample,
                f'sub-{sid}_task-{task}_run-{run:02d}.npy',
            ]
        else:
            fn = [
                fp_version,
                'renamed',
                space,
                lr,
                resample,
                f'sub-{sid}_task-{task}_run-{run:02d}.npy',
            ]
            fn = self.renaming['/'.join(fn)].split('/')

        dm = self.dl_dset.get(fn, on_missing='raise').astype(np.float64)

        return dm

    def load_confounds(self, sid, task, run, fp_version=None):
        if fp_version is None:
            fp_version = self.fp_version
        suffix_li = [
            'desc-confounds_timeseries.npy',
            'desc-confounds_timeseries.tsv',
            'desc-mask_timeseries.npy',
        ]
        output = []
        for suffix in suffix_li:
            if self.renaming is None:
                fn = [
                    fp_version,
                    'confounds',
                    f'sub-{sid}_task-{task}_run-{run}_{suffix}',
                ]
            else:
                fn = [
                    fp_version,
                    'renamed_confounds',
                    f'sub-{sid}_task-{task}_run-{run:02d}_{suffix}',
                ]
                fn = self.renaming['/'.join(fn)].split('/')
            o = self.dl_dset.get(fn, on_missing='raise')
            output.append(o)
        return output

    def load_design(self, sid, task, run, fp_version=None):
        if fp_version is None:
            fp_version = self.fp_version
        suffix = 'design.json'
        if self.renaming is None:
            fn = [
                fp_version,
                'design',
                f'sub-{sid}_task-{task}_run-{run:02d}_{suffix}',
            ]
        else:
            fn = [
                fp_version,
                'renamed_design',
                f'sub-{sid}_task-{task}_run-{run:02d}_{suffix}',
            ]
            fn = self.renaming['/'.join(fn)].split('/')
        output = self.dl_dset.get(fn, on_missing='raise')
        return output

    def load_contrasts(
        self,
        sid,
        task,
        run,
        lr,
        kind='t',
        space=None,
        resample=None,
        prep=None,
        force_volume=False,
        fp_version=None,
    ):
        if force_volume:
            space_kind = 'volume'
        else:
            space_kind = guess_surface_volume(space, resample, lr)
        if space is None:
            space = {
                'surface': self.surface_space,
                'volume': self.volume_space,
            }[space_kind]
        if resample is None:
            resample = {
                'surface': self.surface_resample,
                'volume': self.volume_resample,
            }[space_kind]
        if prep is None:
            prep = self.prep

        if fp_version is None:
            fp_version = self.fp_version
        suffix = f'{kind}.npy'
        fn = [
            fp_version,
            'contrasts',
            space,
            f'{lr}-cerebrum',
            resample,
            prep,
            f'sub-{sid}_task-{task}_run-{run:02d}_{suffix}',
        ]
        output = self.dl_dset.get(fn, on_missing='raise')
        return output

    def get_data(
        self,
        sid,
        task,
        run,
        lr,
        space=None,
        resample=None,
        prep=None,
        fp_version=None,
        force_volume=False,
        prep_kwargs=None,
        slicer=None,
    ):
        if isinstance(run, (tuple, list)):
            ret = [
                self.get_data(
                    sid,
                    task,
                    run_,
                    lr,
                    space,
                    resample,
                    prep,
                    fp_version,
                    force_volume,
                    prep_kwargs,
                )
                for run_ in run
            ]
            if isinstance(ret[0], tuple):
                n = len(ret[0])
                ret = tuple(
                    [
                        np.concatenate([ret_[i] for ret_ in ret], axis=0)
                        for i in range(n)
                    ]
                )
            elif isinstance(ret[0], np.ndarray):
                ret = np.concatenate(ret, axis=0)
            return ret

        if force_volume:
            space_kind = 'volume'
        else:
            space_kind = guess_surface_volume(space, resample, lr)
        if space is None:
            space = {
                'surface': self.surface_space,
                'volume': self.volume_space,
            }[space_kind]
        if resample is None:
            resample = {
                'surface': self.surface_resample,
                'volume': self.volume_resample,
            }[space_kind]
        if prep is None:
            prep = self.prep
        if fp_version is None:
            fp_version = self.fp_version
        if slicer is None:
            slicer = getattr(self, 'slicer', None)

        dm = self.load_data(sid, task, run, lr, space, resample, fp_version)
        confounds = self.load_confounds(sid, task, run, fp_version)
        if space_kind == 'surface':
            if lr == 'lr':
                cortical_mask = np.concatenate(
                    [get_mask(lr_, space) for lr_ in 'lr'], axis=0
                )
            else:
                cortical_mask = get_mask(lr, space)
        else:
            cortical_mask = None
        if isinstance(prep, str):
            prep = get_prep(prep, **(prep_kwargs if prep_kwargs is not None else {}))
        if slicer is not None:
            dm = slicer(dm, task, run)
            confounds = [slicer(c, task, run) for c in confounds]
        dm = prep(dm, confounds, cortical_mask)
        return dm

    def _get_anatomical_data(self, sid, which, lr, mask, space, fp_version):
        if fp_version is None:
            fp_version = self.fp_version
        if space is None:
            space = self.surface_space
        fn = os.path.join(
            fp_version, 'anatomy', space, 'overlap', which, f'{sid}_{lr}h.npy'
        )
        d = self.dl_dset.get(fn, on_missing='raise')
        d = d.astype(np.float64)
        if mask is not False and mask is not None:
            if isinstance(mask, np.ndarray):
                cortical_mask = mask
            else:
                cortical_mask = get_mask(lr, space)
            d = d[cortical_mask]
        return d

    def morphometry(self, sid, which, lr, mask=True, space=None, fp_version=None):
        return self._get_anatomical_data(
            sid=sid, which=which, lr=lr, mask=mask, space=space, fp_version=fp_version
        )

    def parcellation(self, sid, which, lr, mask=True, space=None, fp_version=None):
        return self._get_anatomical_data(
            sid=sid,
            which=which + '.annot',
            lr=lr,
            mask=mask,
            space=space,
            fp_version=fp_version,
        )


class Bologna(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='20.2.7',
        name='bologna',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.subjects = [f'{_+1:02d}' for _ in range(69)]
        self.tasks = ['rest']


class Forrest(Dataset):
    """The Forrest dataset.

    This dataset contains fMRI responses while participants passively
    viewed the audiovisual movie *Forrest Gump* (1994).

    See Sengupta et al. (2016) in the References for more details.

    References
    ----------
    https://doi.org/10.1038/sdata.2016.93
    https://datasets.datalad.org/?dir=/studyforrest
    https://www.studyforrest.org/
    """

    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='20.2.7',
        name='forrest',
        root_dir=None,
        dl_source='https://gin.g-node.org/neuroboros/forrest',
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.subjects = [
            '01',
            '02',
            '03',
            '04',
            '05',
            '06',
            '09',
            '10',
            '14',
            '15',
            '16',
            '17',
            '18',
            '19',
            '20',
        ]
        self.tasks = [
            "forrest",
            "movielocalizer",
            "objectcategories",
            "retmapccw",
            "retmapclw",
            "retmapcon",
            "retmapexp",
        ]


class Dalmatians(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='20.2.7',
        name='dalmatians',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.subjects = [
            'AB033',
            'AB034',
            'AB035',
            'AB036',
            'AB037',
            'AB038',
            'AB039',
            'AB041',
            'AB042',
            'AB043',
            'AB053',
            'AO003',
            'AO004',
            'AO005',
            'AO006',
            'AO007',
            'AO008',
            'AO009',
            'AO010',
            'AO011',
            'AO027',
            'AV012',
            'AV013',
            'AV014',
            'AV015',
            'AV016',
            'AV017',
            'AV018',
            'AV019',
            'AV022',
            'AV032',
            'VD044',
            'VD045',
            'VD046',
            'VD047',
            'VD048',
            'VD049',
            'VD050',
            'VD051',
            'VD052',
            'VO020',
            'VO021',
            'VO023',
            'VO024',
            'VO025',
            'VO026',
            'VO028',
            'VO029',
            'VO030',
            'VO031',
        ]
        self.tasks = ['dalmatians', 'scrambled']


class SpaceTop(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='23.2.0',
        name='spacetop',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['alignvideo', 'faces']


class CamCAN(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='20.2.7',
        name='camcan',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['bang', 'rest', 'smt']

        mod_dir = os.path.dirname(os.path.realpath(__file__))
        for task in ['bang', 'rest', 'smt']:
            with open(os.path.join(mod_dir, f'camcan_{task}.txt')) as f:
                self.subject_sets[task] = f.read().splitlines()


class ID1000(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='20.2.7',
        name='id1000',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['moviewatching']

        mod_dir = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(mod_dir, f'id1000.txt')) as f:
            self.subjects = f.read().splitlines()


class Raiders(Dataset):
    """The Raiders dataset collected with the Siemens scanner.

    This dataset contains fMRI responses while participants passively
    viewed the audiovisual movie *Raiders of the Lost Ark* (1981).

    See Nastase (2018) in the References for more details.

    References
    ----------
    https://www.proquest.com/docview/2018905893/abstract/F07E13FBBC234A93PQ/1
    """

    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='23.2.0',
        name='raiders',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['raiders', 'actions']
        self.subjects = [
            'sid000005',
            'sid000007',
            'sid000009',
            'sid000010',
            'sid000012',
            'sid000013',
            'sid000020',
            'sid000021',
            'sid000024',
            'sid000029',
            'sid000034',
            'sid000052',
            'sid000102',
            'sid000114',
            'sid000120',
            'sid000134',
            'sid000142',
            'sid000278',
            'sid000416',
            'sid000433',
            'sid000499',
            'sid000522',
            'sid000535',
        ]

    def slicer(self, data, task, run):
        if task == 'raiders':  # stimulus overlap between runs
            if run == 1:
                data = data[:-10]
            elif run == 4:
                data = data[10:]
            elif run in [2, 3]:
                data = data[10:-10]
            else:
                raise ValueError(f"Run {run} not recognized.")
        return data


class PhilipsRaiders(Dataset):
    """The Raiders dataset collected using the Philips scanner.

    This dataset contains fMRI responses while participants passively
    viewed the audiovisual movie *Raiders of the Lost Ark* (1981).

    See Haxby et al. (2011) and Guntupalli et al. (2016) in the References
    for more details.

    References
    ----------
    https://doi.org/10.1016/j.neuron.2011.08.026
    https://doi.org/10.1093/cercor/bhw068
    https://datasets.datalad.org/?dir=/labs/haxby/raiders
    """

    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='20.2.7',
        name='raiders',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )


class Budapest(Dataset):
    """The Budapest dataset.

    This dataset contains fMRI responses while participants passively
    viewed the audiovisual movie *The Grand Budapest Hotel* (2014).

    See Visconti di Oleggio Castello et al. (2020) in the References
    for more details.

    References
    ----------
    https://doi.org/10.1038/s41597-020-00735-4
    https://datasets.datalad.org/?dir=/labs/gobbini/budapest
    https://openneuro.org/datasets/ds003017
    https://doi.org/10.18112/openneuro.ds003017.v1.0.3
    """

    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='23.2.0',
        name='budapest',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['budapest', 'hyperface', 'localizer']
        self.subjects = [
            'sid000005',
            'sid000007',
            'sid000009',
            'sid000010',
            'sid000013',
            'sid000020',
            'sid000021',
            'sid000024',
            'sid000029',
            'sid000034',
            'sid000052',
            'sid000114',
            'sid000120',
            'sid000134',
            'sid000142',
            'sid000278',
            'sid000416',
            'sid000499',
            'sid000522',
            'sid000535',
            'sid000560',
        ]


class MonkeyKingdom(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='23.2.0',
        name='monkey-kingdom',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['monkey', 'rest', 'localizer', 'language']
        self.subjects = [
            'sid001123',
            'sid001293',
            'sid001294',
            'sid001678',
            'sid001784',
            'sid001830',
            'sid001835',
            'sid001986',
            'sid002015',
            'sid002161',
            'sid002180',
            'sid002317',
            'sid002325',
            'sid002406',
            'sid002414',
            'sid002435',
            'sid002446',
            'sid002449',
            'sid002454',
            'sid002471',
            'sid002499',
            'sid002509',
            'sid002519',
            'sid002570',
        ]


class Life(Dataset):
    """The Life dataset.

    This dataset contains fMRI responses while participants passively
    viewed four audiovisual segments of the *Life* (2009) nature documentary
    ("life"), as well as fMRI responses while performing an attention task
    while viewing video clips of various animals performing different
    actions ("attention").

    During the "attention" session, participants were asked to attend to
    either the behavior ("beh") or taxonomy ("tax") of the animal in the
    video. Participants performed a 1-back repetition detection task for
    either the same behavior or same taxonomy category in two consecutive
    videos. Participants performed 5 runs for each task. In each run, there
    were 20 conditions (5 taxonomic categories x 4 behavioral categories).
    The 5 taxonomic categories were: primates, ungulates, birds, reptiles,
    and insects. The 4 behavioral categories were: eating, fighting,
    running, and swimming.

    The "life" data was collected with TR = 2.5 s and 3 mm isotropic
    voxels. The "attention" data was collected with TR = 2 s and 3 mm
    isotropic voxels.

    See Nastase et al. (2017, 2018) in the References for more details.

    References
    ----------
    https://doi.org/10.1093/cercor/bhx138
    https://doi.org/10.3389/fnins.2018.00316
    https://datasets.datalad.org/?dir=/labs/haxby/life
    https://datasets.datalad.org/?dir=/labs/haxby/attention
    https://openneuro.org/datasets/ds000233
    https://doi.org/10.18112/openneuro.ds000233.v1.0.1
    https://snastase.github.io/data/
    """

    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='20.2.7',
        name='life',
        root_dir=None,
        dl_source='https://gin.g-node.org/neuroboros/life',
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['life', 'tax', 'beh']
        self.subject_sets.update(
            {
                'attention': [
                    'rid000001',
                    'rid000012',
                    'rid000017',
                    'rid000024',
                    'rid000027',
                    'rid000031',
                    'rid000032',
                    'rid000033',
                    'rid000034',
                    'rid000036',
                    'rid000037',
                    'rid000041',
                ],
                'all': [
                    'rid000001',
                    'rid000005',
                    'rid000006',
                    'rid000009',
                    'rid000012',
                    'rid000014',
                    'rid000017',
                    'rid000019',
                    'rid000020',
                    'rid000024',
                    'rid000027',
                    'rid000031',
                    'rid000032',
                    'rid000033',
                    'rid000034',
                    'rid000036',
                    'rid000037',
                    'rid000038',
                    'rid000041',
                ],
            }
        )
        self.subjects = self.subject_sets['all']
        self.contrasts = [
            'primate_eating',
            'primate_fighting',
            'primate_running',
            'primate_swimming',
            'ungulate_eating',
            'ungulate_fighting',
            'ungulate_running',
            'ungulate_swimming',
            'bird_eating',
            'bird_fighting',
            'bird_running',
            'bird_swimming',
            'reptile_eating',
            'reptile_fighting',
            'reptile_running',
            'reptile_swimming',
            'insect_eating',
            'insect_fighting',
            'insect_running',
            'insect_swimming',
            'primate',
            'ungulate',
            'bird',
            'reptile',
            'insect',
            'eating',
            'fighting',
            'running',
            'swimming',
        ]

    def slicer(self, data, task, run):
        """
        The duration of the 4 video stimuli for the "life" data are: 15:21.32,
        14:12.32, 15:28.33, 16:55.31. That is, 921.32, 852.32, 928.33, and
        1015.31 seconds, respectively. For runs 1, 2, and 4, the first 8 s of
        the video were skipped. For run 3, the first 8.6 seconds of the video
        were skipped.
        After accounting for the skipped video frames, the length of the
        videos (in TRs) are 365.328, 337.728, 367.892, 402.924, respectively.

        The 4 runs of the "life" data contain 374, 346, 377, and 412 volumes,
        respectively. That is, 935, 865, 942, and 1030 seconds, respectively.

        We remove the last few volumes, which hardly contain any responses
        even after accounting for the hemodynamic delay. One participant only
        has 366 and 338 volumes for the first two runs, respectively.
        We keep the same number of volumes for all participants for the first
        two runs, so that we can use the participant's data.
        """
        if task == 'life':
            end = {1: 366, 2: 338, 3: 370, 4: 405}[run]
            data = data[:end]
        return data


class HBNSSI(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='23.2.0',
        name='hbn-ssi',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['raiders', 'matrix', 'walle', 'afewgoodmen']
        self.subjects = [
            '0031121',
            '0031122',
            '0031123',
            '0031126',
            '0031127',
            '0031130',
            '0031131',
            '0031132',
            '0031133',
            '0031124',
            '0031125',
            '0031128',
            '0031129',
        ]


class Whiplash(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='23.2.0',
        name='whiplash',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['whiplash']


class IBC(Dataset):
    def __init__(
        self,
        space=['onavg-ico32', 'mni-4mm'],
        resample=['1step_pial_overlap', '1step_linear_overlap'],
        prep='default',
        fp_version='23.2.0',
        name='IBC',
        root_dir=None,
        dl_source=None,
    ):
        super().__init__(
            name,
            dl_source=dl_source,
            root_dir=root_dir,
            space=space,
            resample=resample,
            prep=prep,
            fp_version=fp_version,
        )
        self.tasks = ['raiders', 'tgbu']


datasets = {
    'forrest': Forrest,
    'bologna': Bologna,
    'dalmatians': Dalmatians,
    'spacetop': SpaceTop,
    'camcan': CamCAN,
    'id1000': ID1000,
    'raiders': Raiders,
    'budapest': Budapest,
    'monkeykingdom': MonkeyKingdom,
    'life': Life,
    'hbn-ssi': HBNSSI,
    'whiplash': Whiplash,
    'ibc': IBC,
}


def get_dataset(name):
    if name not in datasets:
        raise ValueError(
            f"Dataset {name} not recognized. Valid datasets are: {datasets.keys()}"
        )
    return datasets[name]()



import numpy as np


def compute_neighbors(faces, nv=None):
    if nv is None:
        nv = faces.max() + 1

    neighbors = [[] for _ in range(nv)]
    for vs in faces:
        for i in range(3):
            for j in range(3):
                if i == j:
                    continue
                if vs[j] in neighbors[vs[i]]:
                    continue
                neighbors[vs[i]].append(vs[j])
    for i in range(nv):
        neighbors[i] = np.array(neighbors[i])

    return neighbors


def compute_neighbor_distances(coords, neighbors):
    nv = coords.shape[0]
    dists = []
    for i in range(nv):
        d = np.linalg.norm(coords[neighbors[i]] - coords[[i]], axis=1)
        dists.append(d)
    return dists


def compute_vertex_normals_sine_weight(coords, faces):
    normals = np.zeros(coords.shape)

    f_coords = coords[faces]
    edges = np.roll(f_coords, 1, axis=1) - f_coords
    del f_coords
    edges /= np.linalg.norm(edges, axis=2, keepdims=True)

    for f, ee in zip(faces, edges):
        normals[f[0]] += np.cross(ee[0], ee[1])
        normals[f[1]] += np.cross(ee[1], ee[2])
        normals[f[2]] += np.cross(ee[2], ee[0])
    normals /= np.linalg.norm(normals, axis=1, keepdims=True)

    return normals


def compute_vertex_normals_equal_weight(coords, faces):
    normals = np.zeros(coords.shape)

    f_coords = coords[faces]
    e01 = f_coords[:, 1, :] - f_coords[:, 0, :]
    e12 = f_coords[:, 2, :] - f_coords[:, 1, :]
    del f_coords

    face_normals = np.cross(e01, e12)
    face_normals /= np.linalg.norm(face_normals, axis=1, keepdims=True)
    for f, n in zip(faces, face_normals):
        normals[f] += n
    normals /= np.linalg.norm(normals, axis=1, keepdims=True)

    return normals


def compute_face_areas(coords, faces):
    e1 = coords[faces[:, 1]] - coords[faces[:, 0]]
    e2 = coords[faces[:, 2]] - coords[faces[:, 0]]
    face_areas = np.linalg.norm(np.cross(e1, e2), axis=1) / 2
    return face_areas


def compute_vertex_areas(coords, faces, face_areas=None):
    if face_areas is None:
        face_areas_per_vertex = compute_face_areas(coords, faces) / 3
    else:
        face_areas_per_vertex = face_areas / 3
    vertex_areas = np.zeros((coords.shape[0],))
    for f, a in zip(faces, face_areas_per_vertex):
        vertex_areas[f] += a
    return vertex_areas



import os

import nibabel as nib
import numpy as np
from scipy.spatial import cKDTree

from .areal import areal
from .barycentric import barycentric, barycentric_vectors, barycentric_weights
from .nnfr import nnfr
from .properties import (
    compute_face_areas,
    compute_neighbor_distances,
    compute_neighbors,
    compute_vertex_areas,
    compute_vertex_normals_equal_weight,
    compute_vertex_normals_sine_weight,
)
from .subdivision import surface_subdivision
from .union import compute_union_sphere
from .voronoi import compute_occupation, native_voronoi, split_triangle, subdivide_edges


class Surface:
    def __init__(self, coords, faces):
        self.coords = coords.astype('float')
        self.faces = faces
        self.nv = self.coords.shape[0]
        self.nf = self.faces.shape[0]

    @property
    def neighbors(self):
        if not hasattr(self, '_neighbors'):
            self._neighbors = compute_neighbors(self.faces, self.nv)
        return self._neighbors

    @property
    def neighbor_distances(self):
        if not hasattr(self, '_neighbor_distances'):
            self._neighbor_distances = compute_neighbor_distances(
                self.coords, self.neighbors
            )
        return self._neighbor_distances

    @property
    def tree(self):
        if not hasattr(self, '_tree'):
            self._tree = cKDTree(self.coords)
        return self._tree

    @property
    def vertex_areas(self):
        if not hasattr(self, '_vertex_areas'):
            self._vertex_areas = compute_vertex_areas(
                self.coords, self.faces, self.face_areas
            )
        return self._vertex_areas

    def vertex_areas_nn(self, n_div=8, t_div=32):
        new_coords, e_mapping, neighbors = subdivide_edges(
            self.coords, self.faces, n_div
        )
        coords = np.concatenate([self.coords, new_coords])
        nn, nnd = native_voronoi(coords, self.faces, e_mapping, neighbors)

        areas = np.zeros((self.nv,))
        face_areas = self.face_areas
        ww = split_triangle(t_div)
        for f_idx, f in enumerate(self.faces):
            a, b, c = sorted(f)
            indices = np.concatenate(
                [e_mapping[(a, b)], e_mapping[(a, c)], e_mapping[(b, c)], [a, b, c]]
            )
            uu = compute_occupation(f_idx, f, coords, indices, nn, nnd, {}, ww)
            for u, m1 in uu.items():
                areas[u] += m1.mean() * face_areas[f_idx]
        return areas

    @property
    def face_areas(self):
        if not hasattr(self, '_face_areas'):
            self._face_areas = compute_face_areas(self.coords, self.faces)
        return self._face_areas

    @property
    def v2f(self):
        if not hasattr(self, '_v2f'):
            self._v2f = [[] for _ in range(self.nv)]
            for i, f in enumerate(self.faces):
                for v in f:
                    self._v2f[v].append(i)
        return self._v2f

    def subdivide(self, n_div):
        coords, faces = surface_subdivision(self.coords, self.faces, n_div)
        if isinstance(self, Sphere):
            norm = np.linalg.norm(coords[self.coords.shape[0] :], axis=1, keepdims=True)
            coords[self.coords.shape[0] :] /= norm
            subdivided = Sphere(coords, faces)
        else:
            subdivided = Surface(coords, faces)
        return subdivided

    @classmethod
    def from_gifti(cls, fn):
        gii = nib.load(fn)
        coords, faces = (_.data for _ in gii.darrays)
        instance = cls(coords, faces)
        return instance

    def to_gifti(self, fn):
        dirname = os.path.dirname(fn)
        if dirname != '':
            os.makedirs(dirname, exist_ok=True)

        if isinstance(self, Sphere):
            coords = self.coords * 100
        else:
            coords = self.coords

        darrays = [
            nib.gifti.GiftiDataArray(
                coords.astype(np.float32),
                intent=nib.nifti1.intent_codes['NIFTI_INTENT_POINTSET'],
                datatype=nib.nifti1.data_type_codes['NIFTI_TYPE_FLOAT32'],
            ),
            nib.gifti.GiftiDataArray(
                self.faces.astype(np.int32),
                intent=nib.nifti1.intent_codes['NIFTI_INTENT_TRIANGLE'],
                datatype=nib.nifti1.data_type_codes['NIFTI_TYPE_INT32'],
            ),
        ]
        gii = nib.gifti.GiftiImage(darrays=darrays)
        nib.save(gii, fn)


class Sphere(Surface):
    def __init__(self, coords, faces):
        super().__init__(coords, faces)
        self.sphericalize()

    def sphericalize(self):
        norm = np.sqrt(np.sum(self.coords**2, axis=1))[:, np.newaxis]
        if not np.allclose(norm, 1):
            self.coords /= norm

    @classmethod
    def from_gifti(cls, fn):
        gii = nib.load(fn)
        coords, faces = (_.data for _ in gii.darrays)
        instance = cls(coords, faces)
        return instance

    @property
    def normals(self):
        if not hasattr(self, '_normals'):
            self._normals = compute_vertex_normals_equal_weight(self.coords, self.faces)
        return self._normals

    @property
    def normals_sine(self):
        if not hasattr(self, '_normals_sine'):
            self._normals_sine = compute_vertex_normals_sine_weight(
                self.coords, self.faces
            )
        return self._normals_sine

    def prepare_barycentric(self):
        if not hasattr(self, 'vecs'):
            f_coords = self.coords[self.faces, :]
            a = f_coords[:, 0, :]
            e1 = f_coords[:, 1, :] - f_coords[:, 0, :]
            e2 = f_coords[:, 2, :] - f_coords[:, 0, :]
            self.vecs = np.stack(
                [a, e1, e2, np.cross(e1, e2), np.cross(e2, a), np.cross(a, e1)], axis=1
            )

    def barycentric(self, coords, **kwargs):
        self.prepare_barycentric()
        return barycentric(
            self.vecs, coords, self.v2f, self.tree, self.faces, self.nv, **kwargs
        )

    def nnfr(self, coords, reverse=True):
        return nnfr(self.coords, coords, reverse=reverse)

    def areal(self, coords1, anat_coords, coords2=None):
        return areal(self, coords1, anat_coords, coords2)

    def areal_highres(self, coords1, anat_coords, coords2=None, n_div=4):
        highres_sphere = self.subdivide(n_div)
        highres_anat = surface_subdivision(anat_coords, self.faces, n_div)[0]
        if coords2 is None:
            mat = highres_sphere.areal(self.coords, highres_anat, coords1)
        else:
            mat = highres_sphere.areal(coords1, highres_anat, coords2)
        return mat

    def dijkstra_subdivision(self, coords1, anat_coords, n_div=4):
        f_indices, weights = barycentric(
            self.vecs,
            coords1,
            self.v2f,
            self.tree,
            self.faces,
            self.nv,
            eps=1e-7,
            return_sparse=False,
        )

    def union(self, to_unite, eps=1e-10):
        if isinstance(to_unite, Sphere):
            coords = to_unite.coords
        elif isinstance(to_unite, np.ndarray):
            coords = to_unite
        else:
            raise TypeError("`to_unite` must be a Surface or ndarray.")
        new_coords, new_faces, indices1, indices2 = compute_union_sphere(
            self, coords, eps=eps
        )
        union_sphere = Sphere(new_coords, new_faces)
        return union_sphere, indices1, indices2
